# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Sequence, Generator
from types import NoneType
from copy import deepcopy

from .digit import Digits
from .trainshape import TrainShape
from .einsumequation import EinsumEquation

def einsum_operation_shape(eq: EinsumEquation) -> TrainShape:
    space = operation_space(eq)
    dims = list(set(dim for op in eq.shapes for dim in op.dims))
    operation_shape = TrainShape(dims, space)
    return operation_shape

def einsum_result_shape(eq: EinsumEquation, op_shape: TrainShape) -> NoneType | TrainShape:
    space = []
    for core in op_shape.digits:
        res_digits = [digit for digit in core if any(digit in dim for dim in eq.result_dims)]
        space.append(res_digits)

    if all(len(digits) != 0 for digits in space):
        return TrainShape(eq.result_dims, space)
    elif any(len(digits) != 0 for digits in space):
        raise ValueError("Inconsistent operation shape: some cores have relevant digits while others do not.")
    return None

def operation_space(eq: EinsumEquation) -> Sequence[Digits]:
    space = eq.shapes[0].digits
    for i in iter_ops(eq, 0):
        op = eq.shapes[i].digits
        tmps = [fuse_operands(space, op),
                fuse_operands(op, space),
                fuse_operands(space, list(reversed(op))),
                fuse_operands(op, list(reversed(space))),
                fuse_operands(list(reversed(op)), space),
                fuse_operands(list(reversed(space)), op),
                fuse_operands(list(reversed(space)), list(reversed(op))),
                fuse_operands(list(reversed(op)), list(reversed(space))),
                ]
        sums = [size_of_space(eq, tmp)-len(tmp) for tmp in tmps]
        min_idx = sums.index(min(sums))
        space = tmps[min_idx]

    space = prune_space(eq, space)
    #for i in range(len(space)):
    #    space[i] = sorted(space[i], key=lambda d: (d.idf, d.idx))
    return space


def prune_space(eq: EinsumEquation, space: Sequence[Digits]) -> list[Digits]:
    """Prune the operation space to only include digits relevant to the equation result."""
    pruned_space = []
    digit_set = set(digit for dim in eq.result_dims for digit in dim)
    tmp = []
    for digits in space:
        if not any(digit in digit_set for digit in digits):
            tmp = fuse_digits(tmp, digits)
        else:
            pruned_space.append(fuse_digits(tmp, digits))
            tmp = []
    if len(tmp) > 0 and len(pruned_space) > 0:
        pruned_space[-1] = fuse_digits(pruned_space[-1], tmp)
    elif len(tmp) > 0:
        pruned_space.append(tmp)
    return pruned_space

def size_of_space(eq: EinsumEquation, space: Sequence[Digits]) -> int:
    """Calculate the size of the operation space based on the equation result."""
    out_sum = 0
    for digits in space:
        prod_dim = 1
        for digit in digits:
            if any(digit in dim for dim in eq.result_dims):
                prod_dim *= digit.base
        out_sum += prod_dim if prod_dim != 1 else 0
    return out_sum

def fuse_operands(ref: Sequence[Digits],
                  op: Sequence[Digits]) -> list[Digits]:
    """Contract the operands according to matching digits."""

    tmp1 = set(digit for digits in ref for digit in digits)
    tmp2 = set(digit for digits in op for digit in digits)
    shared_digits = tmp1 & tmp2 #intersection

    op_idx = 0
    res = list[Digits]()
    for i, ref_digits in enumerate(ref):
        in_res = [share_any_digits(res[i], ref_digits) for i in range(len(res))]
        if len(res) > 0 and any(in_res):
            res.append(ref_digits)
            idx = in_res.index(True)
            lres = len(res)
            for _ in range(lres-idx-1):
                res[idx] = fuse_digits(res[idx], res.pop())
        elif any(digit in shared_digits for digit in ref_digits):
            tmp = ref_digits
            while op_idx < len(op) and not share_any_digits(tmp, op[op_idx]):
                tmp = fuse_digits(tmp, op[op_idx])
                op_idx += 1
            while op_idx < len(op) and share_any_digits(tmp, op[op_idx]):
                tmp = fuse_digits(tmp, op[op_idx])
                op_idx += 1
            res.append(tmp)
        elif 0 < op_idx < len(op):
            tmp = ref_digits
            while op_idx < len(op) and not share_any_digits(tmp, op[op_idx]):
                tmp = fuse_digits(tmp, op[op_idx])
                op_idx += 1
            res.append(tmp)
        else:
            res.append(ref_digits)

    for i in range(op_idx, len(op)):
        res.append(list(op[i]))

    return res

def iter_ops(eq: EinsumEquation, start: int) -> Generator[int]:
    """Generate the indices of operands in the order they should be processed."""
    op_strs = list(deepcopy(op) for op in eq.operands)
    cur_op_str = op_strs.pop(start)
    while len(op_strs) > 0:
        op_str_idx = next_operand_idx(cur_op_str, op_strs)
        op_str = op_strs.pop(op_str_idx)
        cur_op_str += "".join(char for char in op_str if char not in cur_op_str)
        yield eq.operands.index(op_str)

def next_operand_idx(cur_op: str, op_strs: Sequence[str]) -> int:
    """Find the index of the next operand that shares dimensions with the current operand."""
    for i, op in enumerate(op_strs):
        if any(char in cur_op for char in op):
            return i
    raise ValueError("No connected operand found.")

def fuse_digits(set1: Digits, set2: Digits) -> Digits:
    """Fuse two sets of digits, avoiding duplicates based on similarity and mapping."""
    res = list(set1)
    for digit2 in set2:
        if not any(digit1 == digit2 for digit1 in set1):
            res.append(digit2)
    return res

def share_any_digits(set1: Digits, set2: Digits) -> bool:
    """Check if two sets of digits share any digits based on similarity and mapping."""
    for digit1 in set1:
        for digit2 in set2:
            if digit1 == digit2:
                return True
    return False

def share_all_digits(set1: Digits, set2: Digits) -> bool:
    """Check if two sets of digits share any digits based on similarity and mapping."""
    for digit1 in set1:
        for digit2 in set2:
            if digit1 != digit2:
                return False
    return True



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Generator, Sequence
from types import NoneType
from copy import deepcopy
from dataclasses import dataclass

from .backend import ArrayLike
from .localrange import LocalRange
from .trainshape import TrainShape
from .trainbase import TrainBase
from .einsumcontraction import EinsumContraction
from .contractor import ArrayContractor, OptimizeKind, DEFAULT_OPTIMIZER
from .contractorinput import ContractorInput
from .utils import check_operand_shapes, shape_map, get_shapes

@dataclass(kw_only=True, frozen=True)
class EnvironmentData[T: ArrayLike]:
    left: T
    right: T

class Environment:

    optimizer: OptimizeKind
    _contr: EinsumContraction
    _inp: NoneType | ContractorInput = None
    _to_right_exprs: Sequence[ArrayContractor] = []
    _to_left_exprs: Sequence[ArrayContractor] = []

    def __init__(self, 
                 contr: EinsumContraction,
                 optimizer: OptimizeKind=DEFAULT_OPTIMIZER) -> None:
        if contr.result_shape is not None or contr.full_result_shape is not None:
            raise ValueError("InnerGenerator requires an EinsumContraction without result.")

        self.optimizer = deepcopy(optimizer)
        self._contr = deepcopy(contr)

    def __call__[T: ArrayLike](self, *ops: TrainBase[T], expr: bool = False) -> Generator[EnvironmentData[T], LocalRange]:
        gen = self._gen(*ops, expr=expr)
        next(gen) # warm up
        return gen

    def calc_expressions(self, *ops: TrainShape | TrainBase) -> None:
        check_operand_shapes(self._contr.operand_shapes, get_shapes(*ops))
        self._inp = ContractorInput(*ops)
        self._to_right_exprs = [self._to_right_expression(i, *ops) for i in range(len(self._contr))]
        self._to_left_exprs = [self._to_left_expression(i, *ops) for i in range(len(self._contr))]

    # ------------------------------------------------------------------------
    # Contraction generators

    def _gen(self, *ops: TrainBase, expr: bool = False) -> Generator[EnvironmentData, LocalRange]:
        shapes = get_shapes(*ops)
        if self._inp is None:
            self.calc_expressions(*shapes)
        else:
            self._inp.check_operands(*ops)

        if self._inp is None:
            raise RuntimeError("Input cannot be None here.")
        xp, device, dtype = self._inp.infos(*ops)

        data = EnvironmentData(left=xp.zeros(1), right=xp.zeros(1))
        cache = self._init_cache(xp, device, dtype)
        while True:
            lrange = yield data
            idxs = list(range(lrange.begin, lrange.end))
            for i in idxs:
                cache[i+1] = None

            start = False
            for i in range(idxs[0]):
                start = start or cache[i+1] is None
                if start:
                    cache[i+1] = self._contract_to_right(i, cache, ops, expr=expr)

            start = False
            for i in range(len(self._contr)-1, idxs[-1], -1):
                start = start or cache[i+1] is None
                if start:
                    cache[i+1] = self._contract_to_left(i, cache, ops, expr=expr)

            left, right = cache[idxs[0]], cache[idxs[-1]+2]
            if left is None or right is None:
                raise RuntimeError("Environment data could not be computed.")
            data = EnvironmentData(left=left,
                                   right=right)

    def _init_cache(self, xp, device, dtype) -> list[NoneType | ArrayLike]:
        first_ndim = len(self._contr[0].result.left)
        first = xp.ones([1]*first_ndim, dtype=dtype, device=device)

        last_ndim = len(self._contr[-1].result.right)
        last = xp.ones([1]*last_ndim, dtype=dtype, device=device)

        return [first, *[None]*len(self._contr), last]

    def _contract_to_right(self,
                           idx: int,
                           cache: Sequence[NoneType | ArrayLike],
                           ops: Sequence[TrainBase],
                           expr: bool=False) -> ArrayLike:
        if self._inp is None:
            raise RuntimeError("Input cannot be None here.")
        if not expr:
            tns = self._contr[idx].get_data(*ops, idx_map=self._inp.idx_map)
            return self._to_right_exprs[idx](cache[idx], *tns)
        else:
            tns = self._contr[idx].get_data(*ops)
            to_right_expr = self._to_right_expression(idx, *[op.shape for op in ops])
            return to_right_expr(cache[idx], *tns)

    def _contract_to_left(self,
                          idx: int,
                          cache: Sequence[NoneType | ArrayLike],
                          ops: Sequence[TrainBase],
                          expr: bool=False) -> ArrayLike:
        if self._inp is None:
            raise RuntimeError("Input cannot be None here.")
        if not expr:
            tns = self._contr[idx].get_data(*ops, idx_map=self._inp.idx_map)
            return self._to_left_exprs[idx](*tns, cache[idx+2])
        else:
            tns = self._contr[idx].get_data(*ops)
            to_left_expr = self._to_left_expression(idx, *[op.shape for op in ops])
            return to_left_expr(*tns, cache[idx+2])

    # ------------------------------------------------------------------------
    # Expression builders

    def _to_right_expression(
            self,
            idx: int,
            *ops: TrainShape | TrainBase
            ) -> ArrayContractor:
        lcontr = self._contr[idx]

        left = lcontr.result.left
        right = lcontr.result.right
        eq = f"{left}," + ",".join(str(op) for op in lcontr.operands) + f"->{right}"

        smap = shape_map(ops, lcontr)
        tmp = tuple([smap[char] for char in left])

        const, tns = lcontr.get_constants(*ops)
        const = [val+1 for val in const]
        return ArrayContractor(eq, tmp, *tns, optimizer=self.optimizer)

    def _to_left_expression(
            self, 
            idx: int,
            *ops: TrainShape | TrainBase
            ) -> ArrayContractor:
        lcontr = self._contr[idx]

        left = lcontr.result.left
        right = lcontr.result.right
        eq = ",".join(str(op) for op in lcontr.operands) + f",{right}->{left}"

        smap = shape_map(ops, lcontr)
        tmp = tuple([smap[char] for char in right])

        _, tns = lcontr.get_constants(*ops)
        return ArrayContractor(eq, *tns, tmp, optimizer=self.optimizer)



import unittest
from math import prod

from trainsum.digit import Digit
from utils import prime_factorization

class TestDigit(unittest.TestCase):

    def setUp(self) -> None:
        idfs = [0, 1, 2]
        sizes = [6, 10, 15]
        self.digits = []
        for idf, size in zip(idfs, sizes):
            bases = prime_factorization(size)
            factors = [prod(bases[i+1:]) for i in range(len(bases))]
            for i in range(len(bases)):
                self.digits.append(Digit(idf, i, bases[i], factors[i]))

    def test_construction(self):
        self.assertRaises(ValueError, Digit, 0, -1, 2, 1)
        self.assertRaises(ValueError, Digit, 0, 0, -1, 1)
        self.assertRaises(ValueError, Digit, 0, 0, 0, 1)
        self.assertRaises(ValueError, Digit, 0, 0, 1, 1)
        self.assertRaises(ValueError, Digit, 0, 0, 1, 0)
        self.assertRaises(ValueError, Digit, 0, 0, 1, -1)

    def test_eq(self):
        for i in range(len(self.digits)):
            self.assertEqual(self.digits[i], self.digits[i])
            for j in range(i + 1, len(self.digits)):
                self.assertNotEqual(self.digits[i], self.digits[j])

    def test_hash(self):
        self.assertEqual(len(self.digits), len(set(self.digits)))
        self.assertEqual(len(self.digits), len(set(self.digits+self.digits)))

if __name__ == '__main__':
    unittest.main()



import unittest
from itertools import product

from trainsum import TrainSum
from utils import backends

class TestFull(unittest.TestCase):

    def setUp(self):
        self.trainsum = [TrainSum(backend) for backend in backends]
        self.sizes = [120, 256, 280, 1024]

    def test_full(self):
        vals = [-1.0, 0.0, 1.0]
        for ts, size, val in product(self.trainsum, self.sizes, vals):
            xp = ts.namespace

            shape = ts.trainshape(size)
            exact = xp.full([size], val)
            train = ts.full(shape, val)
            approx = train.to_tensor()
            diff = abs(xp.sum((exact - approx)**2))
            self.assertLess(diff, 1e-7)

if __name__ == "__main__":
    unittest.main()



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Sequence
from math import prod

from .backend import DType, Device, get_index_dtype, ArrayLike, namespace_of_arrays, device
from .digit import Digit
from .dimension import Dimension
from .valueset import ValueSet
from .sequenceof import SequenceOf

class FunctionalCore[T: ArrayLike](SequenceOf[Digit]):
    _left: ValueSet
    _middle: ValueSet
    _right: ValueSet

    @property
    def left(self) -> ValueSet[T]:
        return self._left

    @property
    def middle(self) -> ValueSet[T]:
        return self._middle

    @property
    def right(self) -> ValueSet[T]:
        return self._right

    @property
    def dtype(self) -> DType:
        return self._middle.dtype

    @property
    def device(self) -> Device:
        return self._middle.device

    def __init__(
            self,
            dims: Sequence[Dimension],
            digits: Sequence[Digit],
            left_storage: T,
            right_storage: T) -> None:
        if device(left_storage) != device(right_storage):
            raise ValueError("Left and right storage must be on the same device")

        self._left = ValueSet(left_storage)
        self._right = ValueSet(right_storage)

        xp = namespace_of_arrays(left_storage, right_storage)
        tmp = xp.empty((len(dims), prod(digit.base for digit in digits)),
                       dtype=get_index_dtype(xp),
                       device=device(left_storage))
        self._middle = self._get_middle(dims, digits, tmp)
        super().__init__(digits)

    def _get_middle(self, dims: Sequence[Dimension], digits: Sequence[Digit], storage: T) -> ValueSet[T]:
        xp = namespace_of_arrays(storage)
        int_type = get_index_dtype(xp)
        middle = ValueSet(storage)
        dim_map = {dim.idf: i for i, dim in enumerate(dims)}
        idxs = xp.zeros((len(dims), *[digit.base for digit in digits]), dtype=int_type)
        for i, digit in enumerate(digits):
            dim_idx = dim_map[digit.idf]
            tmp = xp.zeros((len(dims), digit.base), dtype=int_type)
            tmp[dim_idx,:] = xp.asarray([i*digit.factor for i in range(digit.base)], dtype=int_type)
            view = [len(dims),*[1]*i, -1,*[1]*(len(digits)-i-1)]
            idxs += xp.reshape(tmp, view)
        idxs = xp.reshape(idxs, (idxs.shape[0], prod(idxs.shape[1:])))
        middle.add(idxs)
        return middle



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Sequence
from copy import deepcopy
from math import prod
import numpy as np

from .backend import get_index_dtype, ArrayLike, namespace_of_arrays
from .localcontraction import LocalContraction
from .trainshape import TrainShape
from .trainbase import TrainBase

class CoreSlicing[T: ArrayLike]:

    _lcontr: LocalContraction
    _idx_map: dict[int, int]
    _mat_map: dict[tuple[int, int], Sequence[int] | Sequence[slice] | T]
    _bases: Sequence[int]

    def __init__(self, lcontr: LocalContraction, *ops: TrainShape | TrainBase[T]) -> None:
        if len(lcontr.result.middle) == 0:
            raise ValueError("Local contraction without result are and invalid input.")
        self._lcontr = deepcopy(lcontr)
        self._idx_map = _index_map(*ops)
        self._mat_map = _matrix_map(lcontr, *ops)

    def get_matrices(self, comb: int, *ops: TrainBase[T]) -> Sequence[T]:
        mats = []
        for i, (op_idx, core_idx) in enumerate(zip(self._lcontr.train_idxs, self._lcontr.core_idxs)):
            if op_idx in self._idx_map:
                idx = self._idx_map[op_idx]
                cut = self._mat_map[comb, i]
                mats.append(ops[idx].data[core_idx][:,*cut,:])
            else:
                mats.append(self._mat_map[comb, i])
        return mats

    def get_right_shape(self, *ops: TrainBase) -> Sequence[int]:
        shape = {}
        for i, (op, op_idx, core_idx) in enumerate(zip(self._lcontr.operands, self._lcontr.train_idxs, self._lcontr.core_idxs)):
            if op_idx in self._idx_map:
                idx = self._idx_map[op_idx]
                shape[op.right] = ops[idx].data[core_idx].shape[-1]
                shape[op.left] = ops[idx].data[core_idx].shape[0]
            else:
                shape[op.right] = self._mat_map[0, i].shape[-1] # type: ignore
                shape[op.left] = self._mat_map[0, i].shape[0]  # type: ignore
        return [shape[char] for char in self._lcontr.result.right]

def _index_map(*ops: TrainShape | TrainBase) -> dict[int, int]:
    idx_map = {}
    for i in range(len(ops)):
        if not isinstance(ops[i], TrainBase):
            idx_map[i] = len(idx_map)
    return idx_map

def _matrix_map[T: ArrayLike](
        lcontr: LocalContraction,
        *ops: TrainShape | TrainBase[T]
        ) -> dict[tuple[int, int], Sequence[int] | Sequence[slice] | T]:
    rshape = result_shape(lcontr, *ops)
    smap = _slice_map(lcontr, rshape)
    const, tns = lcontr.get_constants(*ops)
    cuts = {}
    for i in range(prod(rshape)):
        for j, (tn, op_str) in enumerate(zip(tns, lcontr.operands)):
            cut = tuple(smap[char][i] if char in smap else slice(None) for char in op_str.middle)
            if j in const: # tn is not a shape
                xp = namespace_of_arrays(tn) # type: ignore
                cuts[i,j] = xp.asarray(tn[:,*cut,:])# type: ignore
            else:
                cuts[i,j] = cut
    return cuts
 
def _slice_map(
        lcontr: LocalContraction,
        rshape: Sequence[int],
        ) -> dict[str, Sequence[slice] | Sequence[int]]:
    tmp = np.zeros((len(rshape), *rshape), dtype=np.uint64)
    ncombs = prod(rshape)
    for i, base in enumerate(rshape):
        ncombs //= base
        for j in range(base):
            cut = (i,) + (slice(None),)*i + (j,) + (slice(None),)*(len(rshape)-i-1)
            tmp[cut] = j
    tmp = np.reshape(tmp, (len(rshape), prod(rshape)))

    smap = {}
    for i, char in enumerate(lcontr.result.middle):
        if char not in smap:
            smap[char] = [int(val) for val in tmp[i,:]]
    return smap

def result_shape(lcontr, *ops: TrainShape | TrainBase) -> Sequence[int]:
    smap = {}
    for op_str, shape in zip(lcontr.operands, lcontr.get_shapes(*ops)):
        for i, char in enumerate(op_str.middle):
            if char not in smap:
                smap[char] = shape[i+1]
    return [smap[char] for char in lcontr.result.middle]



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Callable, overload, Optional

from .backend import ArrayLike, ArrayNamespace
from .trainshape import TrainShape
from .trainbase import TrainBase
from .options import get_options, OptionType
from .crossinterpolation import CrossInterpolation
from .utils import symbol_generator
from .to_train import to_train
from .einsum import einsum

@overload
def construct[T: ArrayLike](shape: TrainShape, func: Callable[[T], T], xp: ArrayNamespace[T], start_idxs: Optional[T] = None, /) -> TrainBase[T]: ...
@overload
def construct[T: ArrayLike](shape: TrainShape, data: T, /) -> TrainBase[T]: ...
# implementation
def construct[T: ArrayLike](
        shape: TrainShape,
        data: T | Callable[[T], T],
        xp: Optional[ArrayNamespace[T]] = None,
        start_idxs: Optional[T] = None, /
        ) -> TrainBase[T]:
    if isinstance(data, Callable):
        if xp is None:
            raise ValueError("Array namespace must be provided when data is a function.")
        opts = get_options(xp, OptionType.CROSS)
        cross = CrossInterpolation(solver=opts.solver, strategy=opts.strategy, eps=opts.eps)
        return cross(xp, data, shape, start_idxs)
    else:
        sgen = symbol_generator()
        chars = "".join(next(sgen) for _ in range(len(shape.dims)))
        eq = f"{chars}->{chars}"
        ref = to_train(shape, data)
        res = einsum(eq, ref, result_shape=shape)
        if not isinstance(res, TrainBase):
            raise RuntimeError("Unexpected result type from EinsumExpression.")
        return res



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from math import cos as mcos
from math import sin as msin
from math import pi

from .backend import ArrayNamespace, ArrayLike
from .uniformgrid import UniformGrid
from .trainshape import TrainShape
from .trainbase import TrainBase
from .full import full

def sin[T: ArrayLike](xp: ArrayNamespace[T], grid: UniformGrid, factor: float, offset: float) -> TrainBase: 
    train = cos(xp, grid, factor, offset+pi/2/abs(factor))
    if factor < 0.0:
        train.set_data(0, -1*train.data[0])
    return train

def cos[T: ArrayLike](xp: ArrayNamespace[T], grid: UniformGrid, factor: float, offset: float) -> TrainBase:
    _check_grid(grid)
    shape = TrainShape(grid.dims[0], [(d,) for d in grid.dims[0]])

    if factor == 0.0:
        return full(xp, shape, 1.0)

    dim = grid.dims[0]
    domain = grid.domains[0]
    spacing = grid.spacings[0]
    off = (domain.lower-offset)/(len(dim))
    
    cores = []
    for digit in dim:
        xi = spacing*digit.factor
        data = xp.zeros((2, digit.base, 2))
        for j in range(digit.base):
            theta = factor * (j*xi + off)
            rot = xp.asarray(_givens_rotation(theta))
            data[:,j,:] = rot
        cores.append(data)

    tmp = xp.ones((1, 2))
    cores[0] = xp.tensordot(tmp, cores[0],
                            axes=([1], [0]))

    tmp = xp.ones((2, 1))
    idx = len(cores[-1].shape)-1
    cores[-1] = xp.tensordot(cores[-1], tmp,
                             axes=([idx], [0]))

    cores[0] = 0.5 * cores[0]
    return TrainBase(shape, cores, copy_data=False)

def _givens_rotation(theta: float) -> list[list[float]]:
    return [[mcos(theta), -msin(theta)],
            [msin(theta),  mcos(theta)]]

def _check_grid(grid: UniformGrid) -> None:
    if grid.ndims != 1:
        raise ValueError("Only 1D uniform grid possible")



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Any, Optional, Type, overload
import h5py
import numpy as np

from .backend import ArrayNamespace, ArrayLike, to_device
from .digit import Digit
from .dimension import Dimension
from .domain import Domain
from .uniformgrid import UniformGrid
from .trainshape import TrainShape
from .trainbase import TrainBase

@overload
def write(group: h5py.Group, obj: Digit) -> None: ...
@overload
def write(group: h5py.Group, obj: Dimension) -> None: ...
@overload
def write(group: h5py.Group, obj: TrainShape) -> None: ...
@overload
def write(group: h5py.Group, obj: Domain) -> None: ...
@overload
def write(group: h5py.Group, obj: UniformGrid) -> None: ...
@overload
def write(group: h5py.Group, obj: TrainBase) -> None: ...
#implementation
def write(group: h5py.Group, obj: Any) -> None:
    if isinstance(obj, Digit):
        group.attrs["idf"] = obj.idf
        group.attrs["idx"] = obj.idx
        group.attrs["base"] = obj.base
        group.attrs["factor"] = obj.factor
    elif isinstance(obj, Dimension):
        group.attrs["idf"] = obj.idf
        group.attrs["prc"] = len(obj)
        group.attrs["bases"] = [digit.base for digit in obj]
    elif isinstance(obj, TrainShape):
        for i, dim in enumerate(obj.dims):
            dgroup = group.create_group(f"dim{i}")
            write(dgroup, dim)
        for i, digits in enumerate(obj.digits):
            cgroup = group.create_group(f"core{i}")
            for j, digit in enumerate(digits):
                dgroup = cgroup.create_group(f"digit{j}")
                write(dgroup, digit)
    elif isinstance(obj, Domain):
        group.attrs["lower"] = obj.lower
        group.attrs["upper"] = obj.upper
    elif isinstance(obj, UniformGrid):
        for i, dim, domain in zip(range(obj.ndims), obj.dims, obj.domains):
            dgroup = group.create_group(f"dim{i}")
            write(dgroup, dim)
            write(dgroup, domain)
    elif isinstance(obj, (TrainBase)):
        sgroup = group.create_group("shape")
        write(sgroup, obj.shape)
        for i, data in enumerate(obj.data):
            group.create_dataset(f"data{i}", data=np.asarray(to_device(data, "cpu")))

@overload
def read(group: h5py.Group, cls: Type[Digit]) -> Digit: ...
@overload
def read(group: h5py.Group, cls: Type[Dimension]) -> Dimension: ...
@overload
def read(group: h5py.Group, cls: Type[TrainShape]) -> TrainShape: ...
@overload
def read(group: h5py.Group, cls: Type[Domain]) -> Domain: ...
@overload
def read(group: h5py.Group, cls: Type[UniformGrid]) -> UniformGrid: ...
@overload
def read[T: ArrayLike](group: h5py.Group, cls: Type[TrainBase[T]], xp: Optional[ArrayNamespace[T]]) -> TrainBase[T]: ...
#implementation
def read(group: h5py.Group, cls: Any, xp: Optional[ArrayNamespace] = None) -> Any:
    if cls == Digit:
        return Digit(int(get_attr(group, "idf")),
                     int(get_attr(group, "idx")),
                     int(get_attr(group, "base")),
                     int(get_attr(group, "factor")))
    elif cls == Dimension:
        return Dimension(get_attr(group, "bases"),
                         idf=int(get_attr(group, "idf")))
    elif cls == TrainShape:
        dims = []
        while f"dim{len(dims)}" in group.keys():
            dgroup = group[f"dim{len(dims)}"]
            assert isinstance(dgroup, h5py.Group)
            dims.append(read(dgroup, Dimension))
        cores = []
        while f"core{len(cores)}" in group.keys():
            cgroup = group[f"core{len(cores)}"]
            assert isinstance(cgroup, h5py.Group)
            digits = []
            while f"digit{len(digits)}" in cgroup.keys():
                dgroup = cgroup[f"digit{len(digits)}"]
                assert isinstance(dgroup, h5py.Group)
                digits.append(read(dgroup, Digit))
            cores.append(digits)
        return TrainShape(dims, cores)
    elif cls == Domain:
        return Domain(float(get_attr(group, "lower")),
                      float(get_attr(group, "upper")))
    elif cls == UniformGrid:
        dims = []
        domains = []
        while f"dim{len(dims)}" in group.keys():
            dgroup = group[f"dim{len(dims)}"]
            assert isinstance(dgroup, h5py.Group)
            dims.append(read(dgroup, Dimension))
            domains.append(read(dgroup, Domain))
        return UniformGrid(dims, domains)
    elif cls == TrainBase:
        if xp is None:
            raise ValueError("Array namespace must be provided to read TrainBase.")
        sgroup = group["shape"]
        assert isinstance(sgroup, h5py.Group)
        shape = read(sgroup, TrainShape)
        data = []
        while f"data{len(data)}" in group.keys():
            dataset = group[f"data{len(data)}"]
            assert isinstance(dataset, h5py.Dataset)
            data.append(xp.asarray(np.asarray(dataset)))
        return TrainBase(shape, data)

    raise ValueError("Invalid class.")

def get_attr(group: h5py.Group, name: str) -> Any:
    return group.attrs[name]



import unittest
from copy import deepcopy
import opt_einsum as oe

from trainsum import TrainSum
from trainsum.typing import Dimension, TensorTrain, TrainShape
from utils import backends

class TestEinsum(unittest.TestCase):

    def setUp(self):
        self.trainsum = [TrainSum(backend) for backend in backends]

    def poly_val(self, ts: TrainSum, dim: Dimension, coeffs: list[float]) -> TensorTrain:
        domain = ts.domain(-1, 1)
        grid = ts.uniform_grid(dim, domain)
        vec = ts.polyval(grid, coeffs, 0.0)
        return vec

    def compare_vals(self, ts:TrainSum, msg: str, eq_str: str, guess: float | TensorTrain, *ops: TensorTrain) -> None:
        xp = ts.namespace
        if not isinstance(guess, (int, float)):
            res = guess.to_tensor()
        else:
            res = guess
        exact = oe.contract(eq_str, *[op.to_tensor() for op in ops])
        diff = xp.sum((res - exact)**2)
        self.assertLess(diff, 1e-6, f"{msg} relative error {diff:.2E} is too large")

    def compare(self, ts: TrainSum, eq_str: str, *ops: TensorTrain, result: None | TrainShape = None) -> None:
        shapes = [op.shape for op in ops]

        if eq_str.split("->")[1]  == "":
            expr = ts.einsum_expression(eq_str, *shapes)
            res = expr(*ops)
            self.compare_vals(ts, "full", eq_str, res, *ops)
            return

        max_rank, cutoff, ncores, nsweeps = 5, 1e-10, 2, 1

        with ts.exact():
            exact_expr = ts.einsum_expression(eq_str, *shapes)

        with ts.decomposition(max_rank=max_rank, cutoff=cutoff, ncores=ncores):
            decomp_expr = ts.einsum_expression(eq_str, *shapes, result_shape=result)

        with ts.variational(max_rank=max_rank, cutoff=cutoff, ncores=ncores, nsweeps=nsweeps):
            var_expr = ts.einsum_expression(eq_str, *shapes, result_shape=result)

        with ts.evaluation():
            ev_expr = ts.evaluate_expression(eq_str, *shapes)

        exact_train = exact_expr(*ops)
        self.compare_vals(ts, "exact", eq_str, exact_train, *ops)
        
        decomp_train = decomp_expr(*ops)
        self.compare_vals(ts, "decomp", eq_str, decomp_train, *ops)

        var_train = var_expr(*ops)
        self.compare_vals(ts, "variational", eq_str, var_train, *ops)
        if not isinstance(var_train, TensorTrain):
            raise ValueError("variational expression should return a TensorTrain")

        xp = ts.namespace
        dims = var_train.shape.dims
        idxs = xp.zeros((len(dims), *[dim.size() for dim in dims]), dtype=ts.index_type)
        for i, dim in enumerate(dims):
            cut = [xp.newaxis] * i + [slice(None)] + [xp.newaxis] * (len(dims)-i-1)
            idxs[i,...] += xp.arange(dim.size(), dtype=ts.index_type)[*cut]

        approx = ev_expr(idxs, *ops)
        exact = oe.contract(eq_str, *[op.to_tensor() for op in ops])
        diff = xp.sum((approx - exact)**2) / xp.sum(exact**2)
        self.assertLess(diff, 1e-13, f"eval relative error {diff:.2E} is too large")

    def test_case1(self) -> None:
        for ts in self.trainsum:
            dim1, dim2, dim3 = [ts.dimension(2**6) for _ in range(3)]
            poly1 = self.poly_val(ts, dim1, [1.0, 0.0, 0.0])
            poly2 = self.poly_val(ts, dim2, [1.0, 1.0, 0.0])
            poly3 = self.poly_val(ts, dim3, [0.1, 1.0, 0.0])

            tmp = ts.full(poly2.shape, 1.0)
            op1 = deepcopy(poly1)
            op1.extend(tmp)

            tmp = ts.full(poly1.shape, 1.0)
            op2 = deepcopy(tmp)
            op2.extend(poly2)

            op3 = deepcopy(poly1)
            op3.extend(poly2)
            op3.extend(poly3)

            self.compare(ts, "ab->", op1)
            self.compare(ts, "ab->a", op1)
            self.compare(ts, "ab->b", op1)

            self.compare(ts, "ab,a->", op1, poly3)
            self.compare(ts, "ab,b->", op1, poly3)
            self.compare(ts, "ab,a->a", op1, poly3)
            self.compare(ts, "ab,b->a", op1, poly3)
            self.compare(ts, "ab,a->b", op1, poly3)
            self.compare(ts, "ab,b->b", op1, poly3)
            self.compare(ts, "ab,a->ab", op1, poly3)
            self.compare(ts, "ab,b->ab", op1, poly3)

            self.compare(ts, "ab,bc->ac", op1, op2)
            self.compare(ts, "ab,ab->a", op1, op2)
            self.compare(ts, "ab,ab->ab", op1, op2)
            self.compare(ts, "ab,ab->b", op1, op2)
            self.compare(ts, "ab,ba->a", op1, op2)

            self.compare(ts, "ab,ac->bc", op1, op2)
            self.compare(ts, "ab,cb->ac", op1, op2)

            self.compare(ts, "abc->", op3)
            self.compare(ts, "abc->ab", op3)
            self.compare(ts, "abc,ab->abc", op3, op1)
            self.compare(ts, "abc,ab->ab", op3, op1)
            self.compare(ts, "abc,ab->ac", op3, op1) 
            self.compare(ts, "abc,ab->bc", op3, op1)
            self.compare(ts, "abc,ab->a", op3, op1)
            self.compare(ts, "abc,ab->b", op3, op1)
            self.compare(ts, "abc,ab->c", op3, op1)

            self.compare(ts, "abc,ac->ac", op3, op1)
            self.compare(ts, "abc,ca->ca", op3, op1)
            self.compare(ts, "abc,ad->bcd", op3, op1)
            self.compare(ts, "abc,bd->acd", op3, op1)
            self.compare(ts, "abc,ba->abc", op3, op1)

    def test_case2(self) -> None:
        for ts in self.trainsum:
            xp = ts.namespace
            dim = ts.dimension(120)
            mat = ts.shift(dim, 2)
            shape = TrainShape(dim, [[*dim]])
            data = xp.exp(-xp.linspace(-10, 10, dim.size())**2)
            data = xp.reshape(data, (1, *[d.base for d in dim], 1))
            vec = ts.tensortrain(shape, [data])

            res_shape = ts.trainshape(dim)
            self.compare(ts, "ab,b->a", mat, vec, result=res_shape)
            self.compare(ts, "a,ab->b", vec, mat, result=res_shape)

    def test_case3(self) -> None:
        for ts in self.trainsum:
            dim = ts.dimension(2**8)

            mat = ts.shift(dim, 2)
            vec = self.poly_val(ts, dim, [1.0, 0.0, 0.0])

            self.compare(ts, "ab,b->a", mat, vec)
            self.compare(ts, "a,ab->b", vec, mat)

    def test_case4(self) -> None:
        for ts in self.trainsum:
            dim1 = ts.dimension(2**5)
            dim2 = ts.dimension(2**5)

            op1 = self.poly_val(ts, dim1, [1.0, 0.0, 0.0])
            op2 = self.poly_val(ts, dim2, [1.0, 1.0, 0.0, 0.0])

            cshapes  = [(d,) for d in dim1[:-1]]
            cshapes += [(dim1[-1], dim2[0])]
            cshapes += [(d,) for d in dim2[1:]]
            shape = TrainShape([dim1, dim2], cshapes)
            data  = [dat for dat in op1.cores[:-1]]
            data += [oe.contract("iaj,mbn->iabn", op1.cores[-1], op2.cores[0])]
            data += [dat for dat in op2.cores[1:]]
            op3 = ts.tensortrain(shape, data)

            self.compare(ts, "a,ab,b->ab", op1, op3, op2)

if __name__ == "__main__":
    unittest.main()



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Protocol
from dataclasses import dataclass
from .backend import ArrayLike

@dataclass(kw_only=True)
class MatrixDecompositionResult[T: ArrayLike]:
    #: Left matrix of the decomposition.
    left: T
    #: Right matrix of the decomposition.
    right: T

class MatrixDecomposition[T: ArrayLike, S: MatrixDecompositionResult](Protocol):

    def right(self, mat: T) -> S: ...
    """Decompose input matrix into some left matrix and a right orthonormal matrix"""

    def right_shape(self, shape: tuple[int, int]) -> tuple[tuple[int, int], tuple[int, int]]: ...
    """Given the shape of an input matrix, return the shape of the output matrices"""

    def left(self, mat: T) -> S: ...
    """Decompose input matrix into a left orthonormal matrix and some right matrix"""

    def left_shape(self, shape: tuple[int, int]) -> tuple[tuple[int, int], tuple[int, int]]: ...
    """Given the shape of an input matrix, return the shape of the output matrices"""



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from .backend import ArrayLike, namespace_of_arrays
import array_api_compat

class LstsqSolver:
    def __call__[T: ArrayLike](self, A: T, b: T) -> T:
        if array_api_compat.is_cupy_array(A):
            import cupy as cp
            return cp.linalg.lstsq(A, b)[0] # type: ignore

        xp = namespace_of_arrays(A, b)
        if not (hasattr(xp, "linalg")) or not (hasattr(xp.linalg, "lstsq")):
            raise NotImplementedError(
                f"Method linalg.lstsq is not implemented for backend {xp}.")
        return xp.linalg.lstsq(A, b)[0] # type: ignore



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from .backend import ArrayLike
from .dimension import Dimension
from .trainshape import TrainShape, change_dims
from .trainbase import TrainBase

def fuse[T: ArrayLike](train1: TrainBase[T], train2: TrainBase[T]) -> TrainBase[T]:
    dims1 = [Dimension([d.base for d in dim]) for dim in train1.shape.dims]
    shape1 = change_dims(train1.shape, dims1)

    dims2 = [Dimension([d.base for d in dim]) for dim in train2.shape.dims]
    shape2 = change_dims(train2.shape, dims2)

    dims = [*dims1, *dims2]
    digits = [*shape1.digits, *shape2.digits]

    shape = TrainShape(dims, digits)
    data = [*train1.data, *train2.data]
    return TrainBase(shape, data)



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from .backend import ArrayLike, namespace_of_arrays

class EigHSolver:
    def __call__[T: ArrayLike](self, mat: T) -> tuple[T, T]:
        xp = namespace_of_arrays(mat)
        if not hasattr(xp, "linalg"):
            raise NotImplementedError(
                f"Extension linalg is missing from namespace {xp}.")
        vals, vecs = xp.linalg.eigh(mat)
        return vals, vecs




from typing import Sequence
import unittest
from itertools import product

from trainsum import TrainSum
from trainsum.typing import UniformGrid
from utils import backends

class TestLinsolver(unittest.TestCase):

    def setUp(self):
        self.trainsum = [TrainSum(backend) for backend in backends]
        self.sizes = [(1024,)]

    def get_grid(self, ts, sizes: Sequence[int], lower: float, upper: float):
        dims = [ts.dimension(size) for size in sizes]
        domains = [ts.domain(lower, upper) for _ in sizes]
        return ts.uniform_grid(dims, domains)

    def get_idxs(self, ts, grid: UniformGrid):
        xp = ts.namespace
        idxs = xp.zeros([len(grid.dims), *[dim.size() for dim in grid.dims]],
                        dtype=ts.index_type)
        for i, dim in enumerate(grid.dims):
            cut = (*(xp.newaxis,) * i, slice(None), *(xp.newaxis,) * (len(grid.dims) - i - 1))
            idxs[i] += xp.arange(dim.size(), dtype=ts.index_type)[cut]
        return idxs

    def test_solve(self) -> None:
        for ts, sizes in product(self.trainsum, self.sizes):
            xp = ts.namespace
            grid = self.get_grid(ts, sizes, -20.0, 20.0)
            idxs = self.get_idxs(ts, grid)
            coords = grid.to_coords(idxs)

            shape = ts.trainshape(*sizes)
            data = 2*xp.exp(-xp.sum(coords**2, axis=0)) + 1
            train = ts.tensortrain(shape, data)
            lmap = ts.linear_map("i,i->i", train, train.shape)
            rhs = ts.full(shape, 1.0)
            guess = ts.full(shape, 1.0)

            solver = ts.gmres(subspace=3, nsteps=1, eps=1e-8)

            decomp = ts.svdecomposition(max_rank=15, cutoff=1e-15)
            strat = ts.sweeping_strategy(ncores=2, nsweeps=20)
            linsolver = ts.linsolver(rhs,
                                     lmap,
                                     solver=solver,
                                     decomposition=decomp,
                                     strategy=strat,
                                     method="dmrg")
            guess = linsolver(guess)
            diff = xp.sum((train.to_tensor() * guess.to_tensor() - rhs.to_tensor())**2)
            self.assertLess(diff, 1e-7)

            decomp = ts.svdecomposition(max_rank=2, cutoff=1e-15)
            strat = ts.sweeping_strategy(ncores=1, nsweeps=10)

            linsolver = ts.linsolver(rhs,
                                     lmap,
                                     solver=solver,
                                     decomposition=decomp,
                                     strategy=strat,
                                     method="amen")
            guess = linsolver(guess)
            diff = xp.sum((train.to_tensor() * guess.to_tensor() - rhs.to_tensor())**2)
            self.assertLess(diff, 1e-7)

if __name__ == "__main__":
    unittest.main()



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Callable, Sequence, Literal
from types import NoneType
from copy import deepcopy
import opt_einsum as oe
#from opt_einsum.typing import OptimizeKind
from .backend import ArrayLike, shape

OptimizeKind = Literal["optimal", "dp", "greedy", "random-greedy", "random-greedy-128", "branch-all", "branch-2", "auto", "auto-hq"]
DEFAULT_OPTIMIZER: OptimizeKind = "greedy"

class ArrayContractor[T: ArrayLike]:

    _ops: list[NoneType | ArrayLike]
    _shapes: Sequence[Sequence[int]]
    _opt: OptimizeKind
    _idxs: Sequence[int]
    _expr: Callable[..., ArrayLike]

    def __init__(
            self,
            eq: str,
            *ops: Sequence[int] | T,
            optimizer: OptimizeKind = DEFAULT_OPTIMIZER) -> None:
        self._shapes = []
        self._ops = []
        self._idxs = []
        for i, op in enumerate(ops):
            if isinstance(op, Sequence):
                self._shapes.append(op)
                self._ops.append(None)
                self._idxs.append(i)
            else:
                self._shapes.append(shape(op))
                self._ops.append(deepcopy(op))
        if len(self._idxs) == 0:
            res = oe.contract(eq, *self._ops, optimize=optimizer)
            self._expr = lambda: res
        else:
            self._expr = oe.contract_expression(
                eq,
                *self._shapes,
                optimize=optimizer)
            #print(str(self._expr).split("\n")[0])

    def __call__(self, *ops: T) -> T:
        if len(ops) != len(self._idxs):
            raise ValueError("Number of provided operands does not match the number required operands.")
        if len(self._idxs) == 0:
            return self._expr() # type: ignore
        for idx, op in zip(self._idxs, ops):
            self._ops[idx] = op
        return self._expr(*self._ops) # type: ignore



from typing import Sequence, Any
import unittest
from itertools import product
from math import prod

from trainsum import TrainSum
from trainsum.trainshape import TrainShape
from trainsum.typing import Dimension, Digit
from utils import backends

class TestTrainShape(unittest.TestCase):

    def setUp(self):
        self.trainsum = [TrainSum(backend) for backend in backends]
        self.sizes = [(120,), (256,), (120, 7, 10), (17, 256, 800)]
        self.modes = ("block", "interleaved")

    def block_digits(self, dims: Sequence[Dimension]) -> Sequence[Sequence[Digit]]:
        return [(d,) for dim in dims for d in dim]

    def interleaved_digits(self, dims: Sequence[Dimension]) -> Sequence[Sequence[Digit]]:
        digits = []
        for i in range(max(len(dim) for dim in dims)):
            digits.append([])
            for dim in dims:
                if i < len(dim):
                    digits[-1].append(dim[i])
        return digits

    def exact_ranks(self, shape: Sequence[Sequence[Digit]]) -> Sequence[int]:
        left = [prod(d.base for d in shape[0])]
        for i in range(1, len(shape)-1):
            left.append(left[-1] * prod(d.base for d in shape[i]))

        right = [prod(d.base for d in shape[-1])]
        for i in range(1, len(shape)-1)[::-1]:
            right.append(right[-1] * prod(d.base for d in shape[i]))
        right.reverse()

        return [min(l, r) for l, r in zip(left, right)]

    def get_shapes(self, ts: Any, dims: Sequence[Dimension], mode: str) -> tuple[TrainShape, Sequence[Sequence[Digit]]]:
        if mode == "block":
            return ts.trainshape(*dims, mode=mode), self.block_digits(dims)
        elif mode == "interleaved":
            return ts.trainshape(*dims, mode=mode), self.interleaved_digits(dims)
        raise ValueError(f"Invalid mode: {mode}")

    # -------------------------------------
    # tests

    def test_construction(self):
        for ts, sizes, mode in product(self.trainsum, self.sizes, self.modes):
            dims = [ts.dimension(size) for size in sizes]
            shape, dgts = self.get_shapes(ts, dims, mode)
            self.assertRaises(ValueError, ts.trainshape, *dims, mode="invalid_mode")

            tmp = [ts.dimension(size) for size in sizes]
            self.assertRaises(ValueError, ts.trainshape, *dims, digits=dgts[:-1]) # not all digits
            self.assertRaises(ValueError, ts.trainshape, *tmp, digits=dgts) # digits don't match dims

            self.assertEqual(len(shape), len(dgts))
            self.assertEqual(shape.dims, dims)
            self.assertTrue(all(d1==d2 for d1, d2 in zip(dgts, shape.digits)))
            self.assertTrue(all(d1==d2 for d1, d2 in zip(dgts, ts.trainshape(*dims, digits=dgts).digits)))
    
    def test_middle(self):
        for ts, sizes, mode in product(self.trainsum, self.sizes, self.modes):
            dims = [ts.dimension(size) for size in sizes]
            shape, dgts = self.get_shapes(ts, dims, mode)
            for i, dgts in enumerate(shape.digits):
                self.assertEqual(shape.middle(i), [d.base for d in dgts])

    def test_exact_ranks(self):
        for ts, sizes, mode in product(self.trainsum, self.sizes, self.modes):
            dims = [ts.dimension(size) for size in sizes]
            shape, _ = self.get_shapes(ts, dims, mode)
            branks = self.exact_ranks(shape.digits)
            shape.ranks = None
            self.assertEqual(shape.ranks, branks)
            self.assertEqual([shape.left_rank(i) for i in range(len(shape))], [1, *branks])
            self.assertEqual([shape.right_rank(i) for i in range(len(shape))], [*branks, 1])

    def test_max_ranks(self):
        mrank = 2
        for ts, sizes, mode in product(self.trainsum, self.sizes, self.modes):
            dims = [ts.dimension(size) for size in sizes]
            shape, _ = self.get_shapes(ts, dims, mode)
            branks = [mrank] * (len(shape)-1)
            shape.ranks = mrank
            self.assertEqual(max(shape.ranks), mrank)
            self.assertEqual([shape.left_rank(i) for i in range(len(shape))], [1, *branks])
            self.assertEqual([shape.right_rank(i) for i in range(len(shape))], [*branks, 1])

    def test_reverse(self):
        for ts, sizes, mode in product(self.trainsum, self.sizes, self.modes):
            dims = [ts.dimension(size) for size in sizes]
            shape, _ = self.get_shapes(ts, dims, mode)
            shape_rev = shape.reverse()
            rev_dgts = [list(reversed(digits)) for digits in reversed(shape.digits)]
            self.assertEqual(shape_rev.digits, rev_dgts)
            self.assertEqual(shape_rev.ranks, shape.ranks[::-1])

if __name__ == '__main__':
    unittest.main()



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Protocol
from .backend import ArrayLike

class MatrixEigenvalueDecomposition(Protocol):
    """Protocol for a matrix eigenvalue decomposition."""
    
    def __call__(self, mat: ArrayLike, /) -> tuple[ArrayLike, ArrayLike]:
        """
        Decompose a matrix into its eigenvalues and eigenvectors.
        """
        ...




# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Any, Callable, Optional
from dataclasses import dataclass
from copy import deepcopy
import time
from math import sqrt

from .backend import ArrayLike, Device, DType, namespace_of_arrays, size
from .utils import check_pos, check_non_neg
from .matrixleastsquares import MatrixLeastSquares
from .lstsqsolver import LstsqSolver

class GMRESData[T: ArrayLike]:
    device: Device
    dtype: DType
    subspace: int
    guess: T
    basis: T

    def __init__(self, subspace: int, guess: T) -> None:
        self.device = guess.device
        self.dtype = guess.dtype
        self.subspace = min(subspace, size(guess))
        self.guess = deepcopy(guess)
        xp = namespace_of_arrays(guess)
        self.basis = xp.zeros((self.subspace+1, *guess.shape), 
                              device=self.device, dtype=self.dtype)

@dataclass
class GMRESResult[T: ArrayLike]:
    #: Solution of the linear system.
    array: T
    #: Time taken to compute the solution.
    time: float
    #: Convergence history of the residuals at each step.
    residuals: list[float]

@dataclass
class GMRES:
    """
    GMRES linear solver.
    """

    #: Number of times the subspace is created and solved
    nsteps: int = 10

    #: Size of the Arnoldi basis
    subspace: int = 10

    #: Minimum residual, after which the algorithm is stopped
    eps: float = 1e-8

    #: Least squares solver for solving the matrix in the Arnoldi basis
    solver: MatrixLeastSquares = LstsqSolver()

    def __setattr__(self, name: str, value: Any) -> None:
        if name in ("nsteps", "subspace"):
            check_pos(name, value)
        elif name == "eps":
            check_non_neg(name, value)
        super().__setattr__(name, value)

    def __call__[T: ArrayLike](
            self,
            mat: Callable[[T], T],
            rhs: T,
            guess: Optional[T] = None, /) -> GMRESResult[T]:
        """
        Solve the linear system with mat being a linear operator and rhs being the right hand side of
        the equation. The initial guess can be provided, otherwise it will be initialized to zero.
        """
        xp = namespace_of_arrays(rhs)
        if guess is None:
            guess = xp.zeros_like(rhs)

        self._check_input(rhs, guess)

        stamp = time.time()
        residuals = []
        data = GMRESData[T](self.subspace, guess)

        for _ in range(self.nsteps):
            residual = self._arnoldi_cycle(mat, rhs, data)
            residuals.append(residual)
            if residuals[-1] <= self.eps:
                break

        return GMRESResult(array=data.guess,
                           time=time.time() - stamp,
                           residuals=residuals)

    def _arnoldi_cycle[T: ArrayLike](
            self,
            mat: Callable[[T], T],
            rhs: T,
            data: GMRESData[T]) -> float:
        xp = namespace_of_arrays(rhs)
        subspace, guess, basis, dtype = data.subspace, data.guess, data.basis, data.dtype

        # Work arrays
        hess = xp.zeros((subspace+1, subspace), dtype=dtype)
        cs = xp.zeros(subspace, dtype=dtype)
        sn = xp.zeros(subspace, dtype=dtype)
        g = xp.zeros(subspace+1, dtype=dtype)

        # Initialize with current residual
        r = rhs - mat(guess)
        res_norm = float(xp.sqrt(xp.sum(r*r)))
        if abs(res_norm) < self.eps:
            return res_norm

        basis[0,:] = r / res_norm
        g[0] = res_norm

        idx = 0
        for i in range(subspace):

            # Modified Gram-Schmidt
            self._gram_schmidt(mat, data, hess, i)

            # Apply previous Givens rotations to H column j
            for j in range(i):
                tmp = cs[j] * hess[j,i] + sn[j] * hess[j+1,i]
                hess[j+1,i] = -sn[j] * hess[j,i] + cs[j] * hess[j+1,i]
                hess[j,i] = tmp

            # R = Q*H; zero out subdiagonal entry
            cs[i], sn[i], denom = self._givens(float(hess[i,i]),
                                               float(hess[i+1,i]))
            hess[i,i] = denom
            hess[i+1,i] = 0.0

            # Update g = Q*g
            g[i+1] = -sn[i] * g[i]
            g[i] = cs[i] * g[i]

            # Residual after j+1 steps is |g[j+1]|
            idx += 1
            res_norm = float(g[idx])

            if abs(res_norm) <= self.eps:
                break

        # solve
        if idx == 0:
            return res_norm
        y = self.solver(hess[:idx,:idx], g[:idx])
        guess += xp.tensordot(y, basis[:idx,:], axes=[[0], [0]])
        return abs(float(g[idx]))

    def _gram_schmidt[T: ArrayLike](
            self,
            mat: Callable[[T], T],
            data: GMRESData[T],
            hess: T,
            idx: int) -> None:
        xp = namespace_of_arrays(data.guess)
        basis = data.basis
        vec = mat(basis[idx,:])
        for j in range(idx + 1):
            hess[j,idx] = xp.sum(vec*basis[j,:])
            vec = vec - hess[j,idx] * basis[j,:]
        hess[idx+1,idx] = xp.sqrt(xp.sum(vec*vec))
        if hess[idx+1,idx] > 0.0:
            basis[idx+1,:] = vec / hess[idx+1,idx]
    
    def _givens(self, val1: float, val2: float) -> tuple[float, float, float]:
        denom = sqrt(val1**2 + val2**2)
        if denom == 0.0:
            return 1.0, 0.0, 0.0
        else:
            return val1/denom, val2/denom, denom

    def _check_input(self, rhs: ArrayLike, guess: ArrayLike) -> None:
        if guess.shape != rhs.shape:
            raise ValueError("x0 and b must have the same shape.")
        if guess.device != rhs.device:
            raise ValueError("x0 and b must be on the same device.")
        if guess.dtype != rhs.dtype:
            raise ValueError("x0 and b must have the same dtype.")



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Sequence
from copy import deepcopy
from dataclasses import dataclass

from .dimension import Dimension
from .trainshape import TrainShape, change_dims

@dataclass(frozen=True)
class EinsumEquation:
    result: str
    operands: Sequence[str]

    result_dims: Sequence[Dimension]
    shapes: Sequence[TrainShape]

    def __init__(self, eq: str, *ops: TrainShape) -> None:
        res_str = eq.split("->")[1]
        op_strs = eq.split("->")[0].split(",")
        _check_equation(op_strs, res_str)
        _check_dimensions(op_strs, ops)
        dim_map = _dimension_mapping(op_strs, *ops)
        ops = _einsum_operands(op_strs, ops, dim_map)
        object.__setattr__(self, "result", res_str)
        object.__setattr__(self, "operands", op_strs)
        object.__setattr__(self, "result_dims", [dim_map[char] for char in res_str])
        object.__setattr__(self, "shapes", ops)

    def __str__(self) -> str:
        return ",".join(self.operands) + f"->{self.result}"

def _dimension_mapping(op_strs: Sequence[str], *ops: TrainShape) -> dict[str, Dimension]:
    """Create a mapping from einsum equation characters to dimensions of the operands."""
    dims = {}
    for op_str, op in zip(op_strs, ops):
        for op_char, dim in zip(op_str, op.dims):
            if op_char in dims:
                continue
            dims[op_char] = Dimension([d.base for d in dim])
    return dims

def _einsum_operands(op_strs: Sequence[str], 
                     ops: Sequence[TrainShape],
                     dim_mapping: dict[str, Dimension]) -> tuple[TrainShape]:
    """Create einsum operands with correct dimensions based on the dimension mapping."""
    nops = []
    for op_str, op in zip(op_strs, ops):
        op_dims = [dim_mapping[char] for char in op_str]
        nop = change_dims(op, op_dims)
        nops.append(nop)
    return tuple(nops)

def _check_equation(ops: Sequence[str], res: str) -> None:
    """Check the validity of an einsum equation."""
    if len(ops) == 0:
        raise ValueError("An einsum equation cannot have zero oeprands.")

    op_list = [deepcopy(op) for op in ops]
    ref = set(op_list.pop(0))
    while len(op_list) > 0:
        len_ops = len(op_list)
        for i, op in enumerate(op_list):
            if any(char in ref for char in op):
                ref.update(op_list.pop(i))
                break
        if len_ops == len(op_list):
            raise ValueError("Disjoint einsum graphs are not allowed.")

    if any(char not in ''.join(ops) for char in res):
        raise ValueError("Result dimensions must appear in at least one operand.")

    for op in [*ops, res]:
        if len(set(op)) != len(op):
            raise ValueError(f"Duplicate dimensions in operands are not allowed, {op}")


def _check_dimensions(op_strs: Sequence[str], ops: Sequence[TrainShape]) -> None:
    """Check if the operand dimensions are compatible with the einsum equation and with each other."""
    if len(op_strs) != len(ops):
        raise ValueError(f"Number of provided operands in the equation "\
                         f"({len(op_strs)}) and in the function call "\
                         f"({len(ops)}) do not match")
    dim_map = {}
    for i, (str_op, op) in enumerate(zip(op_strs, ops)):
        if len(str_op) != len(op.dims):
            raise ValueError(f"Operand {i} have a different number of dimensions.")
        for op_char, dim in zip(str_op, op.dims):
            if op_char in dim_map and dim != dim_map[op_char]:
                raise ValueError(f"Dimension mismatch in dimension {op_char}.")
            else:
                dim_map[op_char] = dim




import unittest
from itertools import product

from trainsum import TrainSum
from utils import backends

class TestShift(unittest.TestCase):

    def setUp(self):
        self.trainsum = [TrainSum(backend) for backend in backends]
        self.sizes = [20, 64, 18]

    def shift_matrix(self, xp, size: int, shift: int):
        if shift == 0:
            return xp.eye(size)
        shift_matrix = xp.zeros((size, size))
        sign = 1 if shift > 0 else -1
        shift = abs(shift)
        if sign < 0:
            idx = lambda i: (i+shift, i)
        else:
            idx = lambda i: (i, i+shift)
        for i in range(size):
            if i + shift < size:
                shift_matrix[idx(i)] = 1.0
        return shift_matrix

    def test_shift(self):
        for ts, size in product(self.trainsum, self.sizes):
            xp = ts.namespace

            shifts = list(range(-size+1, size-1))
            circular = [False, True]
            dim = ts.dimension(size)

            for shift, circ in product(shifts, circular):
                exact = self.shift_matrix(xp, dim.size(), shift)
                if circ:
                    tmp = -size+shift if shift >= 0 else size+shift
                    exact += self.shift_matrix(xp, dim.size(), tmp)

                train = ts.shift(dim, shift, circular=circ)
                approx = train.to_tensor()
                diff = abs(xp.sum((exact - approx)**2))
                self.assertLess(diff, 1e-7)

if __name__ == "__main__":
    unittest.main()



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Any
from dataclasses import dataclass
from .backend import ArrayLike, namespace_of_arrays
from .matrixdecomposition import MatrixDecompositionResult
from .utils import check_non_neg, check_pos

@dataclass(kw_only=True)
class SVDecompositionResult[T: ArrayLike](MatrixDecompositionResult[T]):
    #: Left matrix of the decomposition.
    left: T
    #: Right matrix of the decomposition.
    right: T
    #: Singular values.
    singular_values: T

@dataclass(kw_only=True)
class SVDecomposition[T: ArrayLike]:
    """
    Singular value decomposition. The number of singular values to keep is determined by max_rank
    and cutoff. All singular values below cutoff are discarded, and at most max_rank singular values are kept.
    """

    max_rank: int = 50
    cutoff: float = 0.0

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "max_rank" and value is not None:
            check_pos(name, value)
        elif name == "cutoff":
            check_non_neg(name, value)
        super().__setattr__(name, value)

    def right(self, mat: T) -> SVDecompositionResult[T]:
        """Calculate :math:`U \\Sigma V^H` and return :math:`U \\Sigma` and :math:`V^H`."""
        xp = namespace_of_arrays(mat)
        u, s, vh = self._svd(mat)
        u = u * s[xp.newaxis,:]
        return SVDecompositionResult(left=u, right=vh, singular_values=s)

    def left(self, mat: T) -> SVDecompositionResult[T]:
        """Calculate :math:`U \\Sigma V^H` and return :math:`U` and :math:`\\Sigma V^H`."""
        xp = namespace_of_arrays(mat)
        u, s, vh = self._svd(mat)
        vh = s[:,xp.newaxis] * vh
        return SVDecompositionResult(left=u, right=vh, singular_values=s)

    def _svd(self, mat: T) -> tuple[T, T, T]:
        xp = namespace_of_arrays(mat)
        if not hasattr(xp, "linalg"):
            raise NotImplementedError("Linalg extension missing on this backend, implement your own SVDecomposition!.")
        u, s, vh = xp.linalg.svd(mat, full_matrices=False)
        numel = max(1, min(int(xp.sum(s > self.cutoff)), self.max_rank))
        return u[:,:numel], s[:numel], vh[:numel,:]

    def left_shape(self, shape: tuple[int, int]) -> tuple[tuple[int, int], tuple[int, int]]:
        """Calculate the shape of the left function."""
        m, n = shape
        k = min(m, n, self.max_rank)
        return (m, k), (k, n)

    def right_shape(self, shape: tuple[int, int]) -> tuple[tuple[int, int], tuple[int, int]]:
        """Calculate the shape of the right function."""
        m, n = shape
        k = min(m, n, self.max_rank)
        return (m, k), (k, n)

    def __repr__(self) -> str:
        return f"SVDecomposition(max_rank={self.max_rank}, cutoff={self.cutoff})"



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Protocol, Callable, Optional
from .backend import ArrayLike

class LocalLinSolverResult[T: ArrayLike](Protocol):
    """Protocol for the result of a local linear solver."""

    #: The solution array.
    array: T

    #: The residuals of the linear solver.
    residuals: list[float]

class LocalLinSolver[T: LocalLinSolverResult](Protocol):
    """Protocol for a local linear solver."""

    def __call__[S: ArrayLike](
        self,
        mat: Callable[[S], S],
        rhs: S,
        guess: Optional[S] = None, /
        ) -> T:
        """
        Solve a linear problem for a linear map with a right-hand side and an optional initial guess.
        """
        ...



import unittest
from itertools import product

from trainsum import TrainSum
from utils import backends

class TestQftShift(unittest.TestCase):

    def setUp(self):
        self.trainsum = [TrainSum(backend) for backend in backends]
        self.sizes = [120, 256, 280, 1024]

    def test_qftshift(self):
        for ts, size in product(self.trainsum, self.sizes):
            xp = ts.namespace
            dim = ts.dimension(size)
            domain = ts.domain(-1.0, 2.0)
            grid = ts.uniform_grid(dim, domain)

            coeffs = [1.0, 0.2, 0.1]
            offset = 0.2
            x = xp.linspace(domain.lower, domain.upper, dim.size())
            exact = xp.zeros_like(x)
            for i, coeff in enumerate(reversed(coeffs)):
                exact += coeff * (x-offset)**i
            train = ts.polyval(grid, coeffs, offset)

            exact = xp.fft.fftshift(exact) # type: ignore
            train = ts.qftshift(train)
            approx = train.to_tensor()
            diff = abs(xp.sum((exact - approx)**2) / xp.sum(exact**2))
            self.assertLess(diff, 1e-7)

            exact = xp.fft.ifftshift(exact) # type: ignore
            train = ts.iqftshift(train)
            approx = train.to_tensor()
            diff = abs(xp.sum((exact - approx)**2) / xp.sum(exact**2))
            self.assertLess(diff, 1e-7)

if __name__ == "__main__":
    unittest.main()



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from __future__ import annotations
from typing import Sequence
from dataclasses import dataclass

@dataclass(frozen=True, init=False)
class Digit:
    """
    A digit represents a single component of a quantized dimension.
    """

    #-------------------------------------------------------------------------
    #members & properties

    #: The identifier of the dimension this digit belongs to.
    idf: int

    #: The index of the digit within its dimension.
    idx: int

    #: The base of the digit, i.e., the number of possible values it can take.
    base: int

    #: The factor of the digit, i.e., the product of the bases of all subsequent digits.
    factor: int

    #----------------------------------------------------------------------
    #constructor

    def __init__(self, idf: int, idx: int, base: int, factor: int) -> None:
        self._check_input(idx, base, factor)
        object.__setattr__(self, "idf", idf)
        object.__setattr__(self, "base", base)
        object.__setattr__(self, "idx", idx)
        object.__setattr__(self, "factor", factor)

    def _check_input(self, idx: int, base: int, factor: int) -> None:
        if factor < 1:
            raise ValueError(f"Factor must be positive, but got {factor}")
        if base < 2:
            raise ValueError(f"Base must be greater one, but got {base}")
        if idx < 0:
            raise ValueError(f"Index must be positive, but got {idx}")

    #----------------------------------------------------------------------
    #methods

    def __eq__(self, other) -> bool:
        return isinstance(other, Digit)\
               and self.base == other.base\
               and self.idx == other.idx\
               and self.factor == other.factor\
               and self.idf == other.idf

    def __hash__(self) -> int:
        return hash((self.idf, self.base, self.idx, self.factor))

    def __str__(self) -> str:
        return f"Digit(base={self.base},factor={self.factor},idx={self.idx},idf={self.idf})"

Digits = Sequence[Digit]

def digits_similar(d1: Digit, d2: Digit) -> bool:
    return d1.idx == d2.idx and d1.base == d2.base and d1.factor == d2.factor



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from math import prod
from typing import Any
import array_api_compat as api
from array_api_compat import to_device, device
from array_api_compat import size as _size

from .array_namespace import ArrayNamespace, ArrayLike, Device, DType


def get_namespace(obj: Any) -> ArrayNamespace:
    if not api.is_array_api_obj(obj):
        try:
            obj = obj.zeros(1)
        except:
            raise TypeError("Provided object is not a recognized array or namespace.")
    return api.array_namespace(obj) # type: ignore

def namespace_of_arrays[T: ArrayLike](*arrays: T) -> ArrayNamespace[T]:
    return api.array_namespace(*arrays) # type: ignore

def get_index_dtype(xp: ArrayNamespace) -> Any:
    info = xp.__array_namespace_info__()
    dtypes = info.dtypes(kind=None)
    for name in ["uint64", "int64", "uin32", "int32", "uint16", "int16"]:
        if name in dtypes:
            return dtypes[name]
    raise ValueError("No suitable index dtype found")

def shape_size(shape: tuple[int | None, ...]) -> int:
    if any(s is None for s in shape):
        raise ValueError("Shape contains None dimension(s), cannot compute size.")
    return prod(s for s in shape) # type: ignore

def size(array: ArrayLike) -> int:
    val = _size(array)
    if val is None:
        raise ValueError("Array size is unknown (None).")
    return val

def shape(array: ArrayLike) -> tuple[int, ...]:
    shp = array.shape
    if any(s is None for s in shp):
        raise ValueError("Array shape contains None dimension(s).")
    return shp  # type: ignore



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Optional, Sequence, Generator
from types import NoneType
from dataclasses import dataclass
from copy import deepcopy

from .sequenceof import SequenceOf
from .digit import Digit
from .trainshape import TrainShape, change_dims
from .utils import symbol_generator
from .einsumequation import EinsumEquation
from .operationspace import einsum_operation_shape, einsum_result_shape
from .operandstring import OperandString
from .localcontraction import LocalContraction
from .contains import contains

@dataclass(frozen=True)
class EinsumContraction(SequenceOf[LocalContraction]):
    equation: EinsumEquation
    operation_shape: TrainShape
    operand_shapes: Sequence[TrainShape]
    full_result_shape: NoneType | TrainShape
    result_shape: NoneType | TrainShape

    def __init__(self,
                 eq: EinsumEquation,
                 op_shape: Optional[TrainShape] = None,
                 result: Optional[TrainShape] = None) -> None:
        if op_shape is None:
            op_shape = einsum_operation_shape(eq)
        full_res_shape = einsum_result_shape(eq, op_shape)
        op_shape, full_res_shape, res_shape = _get_result(op_shape, full_res_shape, result)

        trains = []
        core_maps = []
        reverse = []
        for train in eq.shapes:
            res = contains(op_shape, train)
            rev, cmap = res.reverse, res.core_idxs
            if len(cmap) == 0:
                raise ValueError("Train cores do not align with operation space.")
            trains.append(train.reverse() if rev else deepcopy(train))
            core_maps.append(cmap)
            reverse.append(rev)

        sgen = symbol_generator()
        digit_sym_map = _digit_symbol_map(eq.operands, trains, sgen)
        train_syms = [_train_symbols(op, digit_sym_map, sgen) for op in trains]

        locs = []
        is_none = full_res_shape is None
        first_left, last_right = next(sgen), next(sgen)
        for i in range(len(op_shape)):
            lcontr = LocalContraction()
            res_str = OperandString()
            perm = []
            zipped = zip(core_maps, trains, train_syms, reverse)
            for train_idx, (cmap, train, syms, rev) in enumerate(zipped):
                idxs = [j for j, idx in enumerate(cmap) if i == idx]
                for j, idx in enumerate(idxs):

                    sym = syms[idx]
                    if i == 0 and idx == 0:
                        sym.left = first_left
                    if i == len(op_shape)-1 and idx == len(train)-1:
                        sym.right = last_right

                    lcontr.add_operand(train_idx,
                                       idx if not rev else len(train)-1-idx,
                                       sym.reverse() if rev else sym)
                    
                    if not is_none:
                        for digit in train.digits[idx]:
                            if digit in full_res_shape.digits[i] and digit_sym_map[digit] not in res_str.middle:
                                res_str.middle += digit_sym_map[digit]
                                perm.append(_digit_index(res_shape, digit))

                    if j == 0 and idx != 0:
                        res_str.left += sym.left
                    if j == len(idxs)-1 and idx != len(train)-1:
                        res_str.right += sym.right

            if i == 0:
                res_str.left = first_left
            if i == len(op_shape)-1:
                res_str.right = last_right

            if not is_none and not full_res_shape.digits[i] is None:
                tmp = deepcopy(perm)
                tmp.sort()
                perm = [tmp.index(p) for p in perm]
                tmp = {idx: i for i, idx in enumerate(perm)}
                res_str.middle = "".join([res_str.middle[tmp[j]] for j in range(len(perm))])

            lcontr.set_result(res_str)
            locs.append(lcontr)
            #print(",".join(str(op) for op in lcontr.operands) + "->" + str(lcontr.result))
            #print("operand idxs:", [val for val in lcontr.train_idxs])
            #print("core idx    :", [val for val in lcontr.core_idxs])
            #print("result digits:", [str(d) for d in lcontr.core])
            #print("-----------------")

        object.__setattr__(self, "equation", eq)
        object.__setattr__(self, "operation_shape", op_shape)
        object.__setattr__(self, "operand_shapes", eq.shapes)
        object.__setattr__(self, "result_shape", res_shape)
        object.__setattr__(self, "full_result_shape", full_res_shape)
        super().__init__(locs)

def get_symbol_generator(einsum_contr: Sequence[LocalContraction]) -> Generator[str]:
    """Get a symbol generator that avoids conflicts with existing symbols in the contraction."""
    char_set = set()
    for lcontr in einsum_contr:
        for op in lcontr.operands:
            char_set.update(str(op))
        char_set.update(str(lcontr.result))
    sym_gen = symbol_generator()
    while True:
        char = next(sym_gen)
        if char in char_set:
            continue
        yield char

def _train_symbols(cores: TrainShape,
                  digit_sym_map: dict[Digit, str],
                  sym_gen: Generator[str]) -> list[OperandString]:
    """Get the symbol representation of a trainshape based on its digits"""
    res = []
    left = next(sym_gen)
    for core in cores.digits:
        op_str = OperandString(left=left, right=next(sym_gen))
        for digit in core:
            op_str.middle += digit_sym_map[digit]
        left = op_str.right
        res.append(op_str)
    return res

def _digit_symbol_map(op_strs: Sequence[str], ops: Sequence[TrainShape], symbol_gen: Generator[str]) -> dict[Digit, str]:
    """Create a mapping of digits to unique symbols"""
    digit_map = dict[Digit, str]()
    tmp_map = {}
    for op_str, op in zip(op_strs, ops):
        for op_char, dim in zip(op_str, op.dims):
            if op_char not in tmp_map:
                tmp_map[op_char] = [next(symbol_gen) for _ in range(len(dim))]
            dim_syms = tmp_map[op_char]
            for digit in dim:
                digit_map[digit] = dim_syms[digit.idx]
    return digit_map

def _get_result(
        op_shape: TrainShape,
        result: NoneType | TrainShape,
        guess_result: NoneType | TrainShape
        ) -> tuple[TrainShape, NoneType | TrainShape, NoneType | TrainShape]:
    """Validate and return the contraction result shape."""
    if result is None and guess_result is None:
        return op_shape, None, None
    elif result is None and guess_result is not None:
        raise ValueError("Cannot provide a guess result to full contraction.")
    elif result is not None and guess_result is None:
        return op_shape, result, deepcopy(result)
    elif result is not None and guess_result is not None:
        guess_result = change_dims(guess_result, result.dims)
        res = contains(result, guess_result)
        rev, cmap = res.reverse, res.core_idxs
        if len(cmap) == 0:
            raise ValueError("Result shape does not align with contraction result.")
        if rev:
            result = result.reverse()
            op_shape = op_shape.reverse()
        return op_shape, result, deepcopy(guess_result)
    raise RuntimeError("Unreachable code in get_result.")

def _digit_index(shape: TrainShape, digit: Digit) -> int:
    """Get the index of a digit in a TrainShape."""
    idx = 0
    for core in shape.digits:
        for d in core:
            if d == digit:
                return idx
            idx += 1
    raise ValueError("Digit not found in TrainShape.")



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Self
from dataclasses import dataclass

@dataclass(kw_only=True, frozen=True)
class LocalRange:
    begin: int
    end: int

    def __init__(self, *, begin: int, end: int) -> None:
        if begin < 0:
            raise ValueError("LocalRange begin must be non-negative.")
        elif end <= begin:
            raise ValueError("LocalRange end must be greater than begin.")
        object.__setattr__(self, "begin", begin)
        object.__setattr__(self, "end", end)

    def __hash__(self) -> int:
        return hash((self.begin, self.end))

    def __lt__(self, other: Self) -> bool:
        if self.begin != other.begin:
            return self.begin < other.begin
        return self.end < other.end



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Literal, Sequence, overload, Any, Type, Callable, Optional
from types import NoneType
from dataclasses import dataclass
import h5py

from .backend import ArrayLike, get_index_dtype, get_namespace
from .trainshape import trainshape as _trainshape, TrainShape
from .trainbase import TrainBase as TrainBase
from .tensortrain import TensorTrain, tensortrain as _tensortrain
from .digit import Digits
from .dimension import Dimension
from .domain import Domain
from .uniformgrid import UniformGrid
from .sweepingstrategy import SweepingStrategy
from .contractor import OptimizeKind, DEFAULT_OPTIMIZER

from .matrixdecomposition import MatrixDecomposition
from .matrixleastsquares import MatrixLeastSquares
from .matrixeigenvaluedecomposition import MatrixEigenvalueDecomposition
from .svdecomposition import SVDecomposition
from .lstsqsolver import LstsqSolver
from .eighsolver import EigHSolver

from .backend import ArrayNamespace

from .full import full as _full
from .exponential import exp as _exp
from .polyval import polyval as _polyval
from .trigonometric import sin as _sin, cos as _cos
from .toeplitz import toeplitz as _toeplitz, shift as _shift
from .discrete_fourier_transform import (
    qft as _qft,
    iqft as _iqft,
    qftshift as _qftshift,
    iqftshift as _iqftshift,
    qftfreq as _qftfreq
)

from .io import write as _write
from .io import read as _read

from .linearmap import LinearMap as _LinearMap
from .lanczos import Lanczos
from .eigsolver import EigSolver as _EigSolver
from .localrange import LocalRange
from .localeigsolver import LocalEigSolver, LocalEigSolverResult

from .gmres import GMRES
from .linsolver import LinSolver as _LinSolver
from .amensolver import AMEnSolver as _AMEnSolver
from .locallinsolver import LocalLinSolver, LocalLinSolverResult

from .add import add
from .einsum import einsum as _einsum
from .einsum import EinsumExpression as _EinsumExpression
from .evaluate import EvaluateExpression as _EvaluationExpression, evaluate as _evaluate
from .min_max import MinMaxResult, min_max as _min_max

from .options import EvaluationOptions, ExactOptions, DecompositionOptions, OptionType, VariationalOptions, CrossOptions, set_options, get_options
from .qrdecomposition import QRDecomposition

class LinearMap[NDArray: ArrayLike]:
    """
    Linear map of qunatics tensor trains defined by an einsum expression.
    It serves as input to the EigSolver and LinSolver classes.
    """

    _map: _LinearMap

    def __init__(
            self,
            eq: str,
            *ops: TrainShape | TensorTrain[NDArray],
            optimizer: OptimizeKind = DEFAULT_OPTIMIZER) -> None:
        ops_ = [op._base if isinstance(op, TensorTrain) else op for op in ops]
        self._map = _LinearMap(eq, *ops_, optimizer=optimizer)

@dataclass(init=False)
class EigSolver[NDArray: ArrayLike, T: LocalEigSolverResult]:
    """
    Variational eigenvalue solver for quantics tensor trains. The eigenvalue equation is\
    defined by the provided linear maps. The DMRG algotihm is using the canonical format
    of tensor trains to map the global eigenvalue problem to a sequence of local problems,
    which are solved by the provided local solver. If the sweeping strategy has ncores>=2
    the decomposition is used to recover the single cores, thereby defining the rank.
    """

    @property
    def solver(self) -> LocalEigSolver[T]:
        """Iterative eigenvalue solver for the local problems."""
        return self._solver.solver
    @solver.setter
    def solver(self, value: LocalEigSolver[T]) -> None:
        self._solver.solver = value

    @property
    def decomposition(self) -> MatrixDecomposition:
        """Matrix decomposition for recovering single cores in multi-core sweeping strategies."""
        return self._solver.decomposition
    @decomposition.setter
    def decomposition(self, value: MatrixDecomposition) -> None:
        self._solver.decomposition = value

    @property
    def strategy(self) -> SweepingStrategy:
        """Sweeping strategy defining the order of the local problems and the number of cores to be optimized simultaneously."""
        return self._solver.strategy
    @strategy.setter
    def strategy(self, value: SweepingStrategy) -> None:
        self._solver.strategy = value

    @property
    def eps(self) -> float:
        """Convergence criterion."""
        return self._solver.eps
    @eps.setter
    def eps(self, value: float) -> None:
        self._solver.eps = value

    _solver: _EigSolver

    def __init__(
            self,
            *maps: LinearMap[NDArray],
            eps: float = 1e-10,
            solver: LocalEigSolver[T],
            decomposition: MatrixDecomposition,
            strategy: SweepingStrategy,
            optimizer: OptimizeKind,
            ) -> None:
        self._solver = _EigSolver(
                solver, *[m._map for m in maps],
                eps=eps,
                decomposition=decomposition,
                strategy=strategy,
                optimizer=optimizer)

    def __call__(
            self,
            guess: TensorTrain[NDArray],
            states: Sequence[TensorTrain[NDArray]] = [], 
            callback: Optional[Callable[[LocalRange, T], bool]] = None,
            ) -> TensorTrain[NDArray]:
        """
        Start with the sweeping DMRG algorithm to solve eigenvalues with an initial guess.
        The solution will be orthogonal to the provided states. The callback is called after
        each local problem is solved and can be used to monitor the convergence of the algorithm.
        Returning True from the callback will stop the algorithm.
        """
        base_states = [state._base for state in states]
        res = self._solver(guess._base, base_states, expr=True, callback=callback)
        return TensorTrain(res, copy_data=False)

@dataclass(init=False)
class LinSolver[NDArray: ArrayLike, T: LocalLinSolverResult]:
    """
    Variationally solve linear equation systems for quantics tensor trains. The equation is\
    defined by the provided linear maps and the right hand side. Using the canonical format, the
    global linear problem is mapped to a sequence of local problems, which are solved by the provided
    local solver. If the sweeping strategy has ncores>=2 the decomposition is used to recover the
    single cores, thereby defining the rank. The method can be specified to use the DMRG algorithm
    or the AMEn algorithm.
    """

    @property
    def solver(self) -> LocalLinSolver[T]:
        """Iterative linear solver for the local problems."""
        return self._solver.solver
    @solver.setter
    def solver(self, value: LocalLinSolver[T]) -> None:
        self._solver.solver = value

    @property
    def decomposition(self) -> MatrixDecomposition:
        """Matrix decomposition for recovering single cores in multi-core sweeping strategies."""
        return self._solver.decomposition
    @decomposition.setter
    def decomposition(self, value: MatrixDecomposition) -> None:
        self._solver.decomposition = value

    @property
    def strategy(self) -> SweepingStrategy:
        """Sweeping strategy defining the order of the local problems and the number of cores to be optimized simultaneously."""
        return self._solver.strategy
    @strategy.setter
    def strategy(self, value: SweepingStrategy) -> None:
        self._solver.strategy = value

    @property
    def eps(self) -> float:
        """Convergence criterion."""
        return self._solver.eps
    @eps.setter
    def eps(self, value: float) -> None:
        self._solver.eps = value

    _solver: Any

    def __init__(
            self,
            rhs: TensorTrain[NDArray],
            *maps: LinearMap[NDArray],
            method: Literal["dmrg", "amen"],
            solver: LocalLinSolver,
            decomposition,
            strategy: SweepingStrategy,
            optimizer: OptimizeKind,
            ) -> None:
        if method == "dmrg":
            self._solver = _LinSolver(
                    solver, rhs._base, *[m._map for m in maps],
                    decomposition=decomposition,
                    strategy=strategy,
                    optimizer=optimizer)
        else:
            self._solver = _AMEnSolver(
                    solver, rhs._base, *[m._map for m in maps],
                    decomposition=decomposition, 
                    strategy=strategy, optimizer=optimizer)

    def __call__(
            self,
            guess: TensorTrain[NDArray],
            callback: Optional[Callable[[LocalRange, T], bool]] = None,
            ) -> TensorTrain[NDArray]:
        """
        Start with the sweeping DMRG or AMEn algorithm to solve the linear equation system with an
        initial guess. The callback is called after each local problem is solved and can be used to
        monitor the convergence of the algorithm. Returning True from the callback will stop the algorithm.
        """
        res = self._solver(guess._base, callback=callback)
        return TensorTrain(res, copy_data=False)

class EinsumExpression:
    """Utility class for typing."""

    _expr: Callable

    def __init__(self, expr: Callable) -> None:
        self._expr = expr

    def __call__[T: TensorTrain](self, *ops: T) -> float | T:
        """Evaluate the einsum expression with the provided tensor trains as operands."""
        base = self._expr(*[op._base for op in ops])
        if isinstance(base, float):
            return base
        elif isinstance(base, TrainBase):
            return type(ops[0])(base, copy_data=False)
        raise RuntimeError("Unexpected return type from einsum expression.")

class EvaluateExpression:
    """Utility class for typing."""

    _expr: Callable

    def __init__(self, expr: Callable) -> None:
        self._expr = expr

    def __call__[T: Any](self, idxs: T, *ops: TensorTrain[T]) -> T:
        """Evaluate the einsum expression with the provided tensor trains as operands."""
        base = self._expr(idxs, *[op._base for op in ops])
        return base

#-------------------------------------------------------------------------------------------------
# Construction wrapper
@dataclass(frozen=True)
class TrainSum[NDArray: Any]:

    #: Array namespace for the underlying array library.
    namespace: ArrayNamespace[NDArray]

    #: Internally used index type.
    index_type: Any

    def __init__(self, namespace: Any) -> None:
        object.__setattr__(self, "namespace", get_namespace(namespace))
        object.__setattr__(self, "index_type", get_index_dtype(self.namespace))

        set_options(self.decomposition(max_rank=25, cutoff=1e-10, ncores=2))
        set_options(self.cross(max_rank=50, eps=1e-10, solver=LstsqSolver()))
        set_options(self.evaluation())

    #-------------------------------------------------------------------------------------------------
    # base wrapper

    def domain(self, lower: float, upper: float) -> Domain:
        """
        Domain, defining a one dimensional interval.
        """
        return Domain(lower, upper)

    @overload
    def dimension(self, size: int, /) -> Dimension: ...
    @overload
    def dimension(self, bases: Sequence[int], /) -> Dimension: ...
    # implementation
    def dimension(self, size: int | Sequence[int], /) -> Dimension:
        """
        Quantized dimension, defined by a sequence of digits. If an integer is provided, it will be 
        factorized into its prime factors to create the digits. Otherwise, the provided sequence of
        integers will be used as the bases of the digits.
        """
        return Dimension(size)

    @overload
    def uniform_grid(self, dim: Dimension, domain: Domain, /) -> UniformGrid: ...
    @overload
    def uniform_grid(self, dims: Sequence[Dimension], domains: Sequence[Domain], /) -> UniformGrid: ...
    # implementation
    def uniform_grid(self, dims: Any, domains: Any, /) -> UniformGrid:
        """
        Uniformly N-dimensional spaced grid.
        """
        return UniformGrid(dims, domains)

    @overload
    def trainshape(self, *dims: Dimension, digits: Sequence[Digits]) -> TrainShape: ...
    @overload
    def trainshape(self, *dims: int | Dimension, mode: Literal["full", "block", "interleaved"] = "block") -> TrainShape: ...
    # implementation
    def trainshape(
            self,
            *dims: int | Dimension,
            mode: Literal["full", "block", "interleaved"] = "block",
            digits: Optional[Sequence[Digits]] = None
            ) -> TrainShape:
        """
        Base class describing the shape of a quantics tensor train.
        """
        return _trainshape(*dims, mode=mode, digits=digits) # type: ignore

    def svdecomposition(self, max_rank: int, cutoff: float = 1e-10) -> SVDecomposition[NDArray]:
        """
        Singular value decomposition for matrices.
        """
        return SVDecomposition(max_rank=max_rank, cutoff=cutoff)

    def qrdecomposition(self) -> QRDecomposition[NDArray]:
        """
        QR decomposition for matrices.
        """
        return QRDecomposition()

    def sweeping_strategy(self, ncores: int, nsweeps: int) -> SweepingStrategy:
        """
        Strategy for sweeping through a tensor train.
        """
        return SweepingStrategy(ncores=ncores, nsweeps=nsweeps)

    #-------------------------------------------------------------------------------------------------
    # construction wrapper

    def full(self, shape: TrainShape, value: float) -> TensorTrain[NDArray]:
        """
        :math:`f(x)=v`.\n
        Construct a quantics tensor train where all entries are set to the given value.
        The resulting TensorTrain will have rank one.
        """
        return TensorTrain(_full(self.namespace, shape, value), copy_data=False)

    def exp(self, grid: UniformGrid, factor: float, offset: float) -> TensorTrain[NDArray]:
        """
        :math:`f(x)=e^{a(x-x_0)}`.\n
        Construct a quantics tensor train representing an exponential function.
        The provided grid has to be one-dimensional. The resulting TensorTrain will have rank one.
        """
        return TensorTrain(_exp(self.namespace, grid, factor, offset), copy_data=False)

    def polyval(self, grid: UniformGrid, coeffs: Sequence[float], offset: float) -> TensorTrain[NDArray]:
        """
        :math:`f(x)=\\sum_i{v_i(x-x_0)^i}`.\n
        Construct a quantics tensor train representing a polynomial on the given grid.
        The provided grid has to be one-dimensional.The resulting TensorTrain will have the rank len(coeffs)+1.
        """
        return TensorTrain(_polyval(self.namespace, grid, coeffs, offset), copy_data=False)

    def sin(self, grid: UniformGrid, factor: float, offset: float) -> TensorTrain[NDArray]:
        """
        :math:`f(x)=\\sin{\\left(a(x-x_0)\\right)}`.\n
        Construct a quantics tensor train representing a sine on the given grid.
        The provided grid has to be one-dimensional.The resulting TensorTrain will have the rank 2.
        """
        return TensorTrain(_sin(self.namespace, grid, factor, offset), copy_data=False)

    def cos(self, grid: UniformGrid, factor: float, offset: float) -> TensorTrain[NDArray]:
        """
        :math:`f(x)=\\cos{\\left(a(x-x_0)\\right)}`.\n
        Construct a quantics tensor train representing a cosine on the given grid.
        The provided grid has to be one-dimensional.The resulting TensorTrain will have the rank 2.
        """
        return TensorTrain(_cos(self.namespace, grid, factor, offset), copy_data=False)

    def shift(self, dim: Dimension, shift: int, circular: bool = False) -> TensorTrain[NDArray]:
        """
        Generic shift matrix. Positive shifts correspond to shifts to the right, negative
        shifts correspond to shifts to the left. If circular is set to True the shift matrix will be
        circular. The resulting tensor train is rank 2.
        """
        return TensorTrain(_shift(self.namespace, dim, shift, circular), copy_data=False)

    def toeplitz(self, dim: Dimension, mode: Literal["full", "lower", "upper", "circular"]) -> TensorTrain[NDArray]:
        """
        Three dimensional Toeplitz tensor. The tensor can be used to create toeplitz matrices
        by contracting the first dimension with a vector. With mode='full' the resulting matrix is a full
        toeplitz matrix, where every diagonal is set one value of the vector. 'lower' and 'upper' correspond
        to lower and upper triangular toeplitz matrices. mode='circular' results in the sum of the lower and
        upper triangular toeplitz matrices. The resulting tensor train is rank 2.
        """
        return TensorTrain(_toeplitz(self.namespace, dim, mode), copy_data=False)

    @overload
    def tensortrain(self, shape: TrainShape, data: Callable[[NDArray], NDArray], start_idxs: Optional[NDArray] = None, /) -> TensorTrain[NDArray]: ...
    @overload
    def tensortrain(self, shape: TrainShape, data: Sequence[NDArray], /) -> TensorTrain[NDArray]: ...
    @overload
    def tensortrain(self, shape: TrainShape, data: NDArray, /) -> TensorTrain[NDArray]: ...
    # implementation
    def tensortrain(
            self,
            shape: TrainShape,
            data: NDArray | Sequence[NDArray] | Callable[[NDArray], NDArray],
            start_idxs: Optional[NDArray] = None,
            /) -> TensorTrain[NDArray]:
        """
        Construct generic quantics tensor trains from some provided data. If data is a function the tensor
        train will be constructed by a cross interpolation (cross context manager). If it is a tensor it will
        be decomposed (einsum context manager). A sequence of tensors are interpreted as the cores of the
        tensor train.
        """
        return _tensortrain(shape, data, start_idxs, self.namespace)# type: ignore

    #-------------------------------------------------------------------------------------------------
    # fourier transform wrapper

    def qft(self, 
            dim: Dimension,
            decomp: MatrixDecomposition = SVDecomposition(max_rank=16)
            ) -> TensorTrain[NDArray]:
        """
        Quantum fourier transformation matrix as a quantics tensor train.
        The qft is constructed approximately using a decomposition algorithm for
        a specified dimension. The decomposition can be specified to control the
        accuracy of the resulting transformation.
        """
        return TensorTrain(_qft(self.namespace, dim, decomp), copy_data=False)

    def iqft(
            self,
            dim: Dimension,
            decomp: MatrixDecomposition = SVDecomposition(max_rank=16)
            ) -> TensorTrain[NDArray]:
        """
        Inverse quantum fourier transformation matrix as a quantics tensor train.
        The construction is the same as for qft, but the result is complex conjugated.
        As for the qft, the decomposition can be specified to control the accuracy of
        the resulting transformation.
        """
        return TensorTrain(_iqft(self.namespace, dim, decomp), copy_data=False)

    def qftshift(self, train: TensorTrain[NDArray], axis: int = 0) -> TensorTrain[NDArray]:
        """
        Shift the zero frequencies to the middle of the spectrum.
        Only works for dimensions with a binary leading bit.
        """
        return TensorTrain(_qftshift(train._base, axis), copy_data=False)

    def iqftshift(self, train: TensorTrain[NDArray], axis: int = 0) -> TensorTrain[NDArray]:
        """
        Inverse of the qftshift.
        Only works for dimensions with a binary leading bit.
        """
        return TensorTrain(_iqftshift(train._base, axis), copy_data=False)

    def qftfreq(self, dim: Dimension, d: float = 1.0) -> TensorTrain[NDArray]:
        """
        Sample frequencies for the quantum fourier transform.
        """
        return TensorTrain(_qftfreq(self.namespace, dim, d), copy_data=False)

    #-------------------------------------------------------------------------------------------------
    # io wrapper

    @overload
    def write(self, group: h5py.Group, obj: UniformGrid) -> None: ...
    @overload
    def write(self, group: h5py.Group, obj: TensorTrain[NDArray]) -> None: ...
    # implementation
    def write(self, group: h5py.Group, obj: Any) -> None:
        """
        Write a quantics tensor train or a uniform grid to a hdf5 group.
        """
        if isinstance(obj, TensorTrain):
            _write(group, obj._base)
        else:
            _write(group, obj)

    @overload
    def read(self, group: h5py.Group, cls: Type[UniformGrid]) -> UniformGrid: ...
    @overload
    def read(self, group: h5py.Group, cls: Type[TensorTrain[NDArray]]) -> TensorTrain[NDArray]: ...
    # implementation
    def read(self, group: h5py.Group, cls: Any) -> Any:
        """
        Read a quantics tensor train or a uniform grid to a hdf5 group.
        """
        if cls == TensorTrain:
            base = _read(group, TrainBase, self.namespace)
            return TensorTrain(base, copy_data=False)
        return _read(group, cls)

    #-------------------------------------------------------------------------------------------------
    # tensorized solver wrapper

    def linear_map(
            self,
            eq: str,
            *ops: TrainShape | TensorTrain[NDArray],
            optimizer: OptimizeKind = DEFAULT_OPTIMIZER,
            ) -> LinearMap[NDArray]:
        """Linear map for variational solvers."""
        return LinearMap(eq, *ops, optimizer=optimizer)

    def lanczos(
            self, *,
            nsteps: int = 1,
            subspace: int = 3,
            eps: float = 1e-8,
            solver: MatrixEigenvalueDecomposition = EigHSolver()
            ) -> Lanczos:
        """
        Lanczos iterative eigenvalue solver.
        """
        return Lanczos(nsteps=nsteps, subspace=subspace, eps=eps, solver=solver)

    def gmres(
            self, *,
            nsteps: int = 10,
            subspace: int = 10,
            eps: float = 1e-8,
            solver: MatrixLeastSquares = LstsqSolver(),
            ) -> GMRES:
        """
        GMRES iterative linear solver.
        """
        return GMRES(nsteps=nsteps, subspace=subspace, eps=eps, solver=solver)

    def eigsolver[T: LocalEigSolverResult](
            self,
            *maps: LinearMap[NDArray],
            eps: float = 1e-10,
            solver: LocalEigSolver[T] = Lanczos(),
            decomposition= SVDecomposition(),
            strategy: SweepingStrategy = SweepingStrategy(ncores=2, nsweeps=10),
            ) -> EigSolver[NDArray, LocalEigSolverResult]:
        """
        Variational eigenvalue solver for quantics tensor trains.
        """
        return EigSolver(
                *maps,
                eps=eps,
                solver=solver,
                decomposition=decomposition,
                strategy=strategy,
                optimizer=DEFAULT_OPTIMIZER)

    def linsolver[T: LocalLinSolverResult](
            self,
            rhs: TensorTrain[NDArray],
            *maps: LinearMap[NDArray],
            method: Literal["dmrg", "amen"] = "dmrg",
            solver: LocalLinSolver[T] = GMRES(),
            decomposition= SVDecomposition(),
            strategy: SweepingStrategy = SweepingStrategy(ncores=2, nsweeps=10),
            optimizer: OptimizeKind = DEFAULT_OPTIMIZER,
            ) -> LinSolver[NDArray, T]:
        """
        Variational linear solver for quantics tensor trains.
        """
        return LinSolver(rhs, *maps, method=method, solver=solver,
                         decomposition=decomposition, strategy=strategy, optimizer=optimizer)


    #-------------------------------------------------------------------------------------------------
    # operation wrapper

    def einsum(
            self,
            eq: str,
            *ops: TensorTrain[NDArray],
            result_shape: NoneType | TrainShape = None
            ) -> float | TensorTrain[NDArray]:
        """
        Perform an einsum operation. The resulting tensor train will always
        be as exact as possible unless a result shape is provided. In that case, the resulting tensor
        train will be decomposed to fit the provided shape. Affected by einsum context manager.
        """
        bases = [op._base for op in ops]
        res = _einsum(eq, *bases, result_shape=result_shape)
        if not isinstance(res, TrainBase):
            return res
        return TensorTrain(res, copy_data=False)
    
    def einsum_expression(
            self,
            eq: str,
            *ops: TrainShape,
            result_shape: NoneType | TrainShape = None
            ) -> EinsumExpression:
        """
        Create an einsum expression to be used with tensor trains. The expression is optimized
        with the ranks of the provided operands. Affected by einsum context manager.
        """
        expr = _EinsumExpression(self.namespace, eq, *ops, result_shape=result_shape)
        return EinsumExpression(expr)

    def evaluate(
            self,
            eq: str,
            idxs: NDArray,
            *ops: TensorTrain[NDArray],
            ) -> NDArray:
        """
        Perform an evaluation at idxs of an einsum expression. Affected by einsum context manager.
        """
        return _evaluate(eq, idxs, *[op._base for op in ops])

    def evaluate_expression(
            self,
            eq: str,
            *ops: TrainShape,
            ) -> EvaluateExpression:
        """
        Create an evaluation expression to be used with tensor trains. The expression is optimized
        with the ranks of the provided operands. Affected by einsum context manager.
        """
        expr = _EvaluationExpression(self.namespace, eq, *ops)
        return EvaluateExpression(expr)

    def min_max(self, train: TensorTrain[NDArray], num: int) -> MinMaxResult[NDArray]:
        """
        Calculate the minimum and maximum values and indices.
        """
        return _min_max(train._base, num)

    def add(
            self,
            train1: TensorTrain[NDArray],
            train2: TensorTrain[NDArray],
            *trains: TensorTrain[NDArray]
            ) -> TensorTrain[NDArray]:
        """
        Add multiple tensor trains. Affected by einsum context manager.
        """
        base = add(train1._base, train2._base, *[train._base for train in trains])
        return TensorTrain(base, copy_data=False)

    #-------------------------------------------------------------------------------------------------
    # default options

    def exact(
            self,
            optimizer: OptimizeKind = DEFAULT_OPTIMIZER) -> ExactOptions:
        """
        Exact einsum operations.
        """
        return ExactOptions(
                namespace=self.namespace,
                optimizer=optimizer)


    @overload
    def decomposition[T: MatrixDecomposition](
            self, *,
            decomposition: T,
            ncores: int = 2,
            optimizer: OptimizeKind = DEFAULT_OPTIMIZER
            ) -> DecompositionOptions[T]: ...
    @overload
    def decomposition(
            self, *,
            max_rank: int,
            cutoff: float = 0.0,
            ncores: int = 2,
            optimizer: OptimizeKind = DEFAULT_OPTIMIZER) -> DecompositionOptions[SVDecomposition[NDArray]]: ...
    def decomposition(
            self, *,
            max_rank: int = 0,
            decomposition: Optional[Any] = None,
            cutoff: float = 0.0,
            ncores: int = 2,
            optimizer: OptimizeKind = DEFAULT_OPTIMIZER
            ) -> DecompositionOptions[Any]:
        """
        Einsum operations based on matrix decompositions. One can provide either a decomposition object
        or paramters for a singular value decomposition (max_rank, cutoff). ncores specifies the number
        of cores to be used for the sweeping strategy.
        """
        if decomposition is None:
            decomposition = SVDecomposition(max_rank=max_rank, cutoff=cutoff)
        strat = SweepingStrategy(ncores=ncores)
        return DecompositionOptions(
                namespace=self.namespace,
                decomposition=decomposition,
                strategy=strat,
                optimizer=optimizer)

    @overload
    def variational[T: MatrixDecomposition](
            self, *,
            decomposition: T,
            ncores: int = 2,
            nsweeps: int = 1,
            optimizer: OptimizeKind = DEFAULT_OPTIMIZER
            ) -> VariationalOptions[T]: ...
    @overload
    def variational(
            self, *,
            max_rank: int,
            cutoff: float = 0.0,
            ncores: int = 2,
            nsweeps: int = 1,
            optimizer: OptimizeKind = DEFAULT_OPTIMIZER) -> VariationalOptions[SVDecomposition[NDArray]]: ...
    def variational(
            self, *,
            max_rank: int = 50,
            cutoff: float = 0.0,
            decomposition: Optional[Any] = None,
            ncores: int = 2,
            nsweeps: int = 1,
            optimizer: OptimizeKind = DEFAULT_OPTIMIZER) -> VariationalOptions:
        """
        Variational einsum operations. One can provide either a decomposition object or
        paramters for a singular value decomposition (max_rank, cutoff). ncores and nsweeps
        specify the sweeping strategy.
        """
        if decomposition is None:
            decomposition = SVDecomposition(max_rank=max_rank, cutoff=cutoff)
        strat = SweepingStrategy(ncores=ncores, nsweeps=nsweeps)
        return VariationalOptions(
                namespace=self.namespace,
                decomposition=decomposition,
                strategy=strat,
                optimizer=optimizer)

    def cross(
            self, *,
            max_rank: int,
            eps: float,
            solver: Optional[MatrixLeastSquares] = None) -> CrossOptions:
        """
        Manager for cross interpolation based operations. max_rank specifies the sweeping
        strategy to be employed. eps and solver are passed to CrossOptions.
        """
        if solver is None:
            solver = LstsqSolver()
        strat = SweepingStrategy(ncores=2, nsweeps=max_rank//2-1)
        return CrossOptions(
                namespace=self.namespace,
                strategy=strat, eps=eps,
                solver=solver)

    def evaluation(
            self, *,
            chunk_size: int = 1024,
            optimizer: OptimizeKind = DEFAULT_OPTIMIZER,
            ) -> EvaluationOptions:
        """
        Manager for evaluate operations.
        """
        return EvaluationOptions(namespace=self.namespace, chunk_size=chunk_size, optimizer=optimizer)


    def set_options(self, options: ExactOptions | DecompositionOptions | VariationalOptions | CrossOptions | EvaluationOptions) -> None:
        """
        Set options globally. The options are stored in a thread local variable and are
        used by the einsum, add, cross and evaluate operations.
        """
        set_options(options)

    def get_options(self, otype: OptionType) -> ExactOptions | DecompositionOptions | VariationalOptions | CrossOptions | EvaluationOptions:
        """
        Get the current options.
        """
        return get_options(self.namespace, otype)



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from typing import Any, Sequence, overload, Optional, Self
from copy import deepcopy

from .backend import ArrayLike, namespace_of_arrays, to_device, device, shape
from .qrdecomposition import QRDecomposition
from .tensordecomposition import TensorDecomposition

from .normalization import Normalization
from .trainshape import TrainShape, transform_index

class TrainBase[T: ArrayLike]:

    _shape: TrainShape
    _data: list[T]
    _device: Any
    _dtype: Any
    _norm: list[Normalization]

    @property
    def device(self) -> Any:
        return self._device
    @device.setter
    def device(self, device: Any) -> None:
        self._device = device
        for i in range(len(self._data)):
            self._data[i] = to_device(self._data[i], device)

    @property
    def dtype(self) -> Any:
        return self._dtype
    @dtype.setter
    def dtype(self, dtype: Any) -> None:
        self._dtype = dtype
        xp = namespace_of_arrays(self._data[0])
        for i in range(len(self._data)):
            self._data[i] = xp.asarray(self._data[i], dtype=dtype, device=self._device)

    @property
    def shape(self) -> TrainShape:
        self._shape.ranks = [shape(tn)[-1] for tn in self._data[:-1]]
        return self._shape

    @property
    def data(self) -> Sequence[T]:
        return self._data

    def __init__(self,
                 shape: TrainShape,
                 data: Sequence[T],
                 norm: Optional[Sequence[Normalization]] = None,
                 copy_data: bool = True) -> None:
        self._check_shape_vs_data(shape, data)
        self._check_dtype_and_device(data)
        self._check_ranks(data)
        if copy_data:
            data = list(deepcopy(core_data) for core_data in data)

        self._shape = deepcopy(shape)
        self._data = list(data)
        self._device = device(data[0])
        self._dtype = data[0].dtype
        if norm is None:
            norm = [Normalization.NONE for _ in data]
        self._norm = list(norm)

    @overload
    def set_data(self,
                 idx: int,
                 data: T, /,
                 norm: Optional[Normalization] = None) -> None: ...
    @overload
    def set_data(self,
                 cut: slice,
                 data: Sequence[T], /,
                 norm: Optional[Sequence[Normalization]] = None) -> None: ...
    # implementation
    def set_data(self,
                 idx: int | slice,
                 data: T | Sequence[T], /,
                 norm: Optional[Normalization | Sequence[Normalization]] = None
                 ) -> None:
        if isinstance(idx, int)\
        and isinstance(data, type(self._data[0]))\
        and isinstance(norm, (Normalization, type(None))):
            if idx < 0 or idx >= len(self._data):
                raise IndexError("Core index out of range.")
            if data.shape != self._data[idx].shape:
                raise ValueError("New data shape does not match existing core shape.")
            self._data[idx] = data
            if norm is None:
                norm = Normalization.NONE
            self._norm[idx] = norm
        elif not isinstance(idx, int)\
        and not isinstance(data, type(self._data[0]))\
        and not isinstance(norm, Normalization):
            self._data[idx] = list(data) # type: ignore
            for i in range(idx.start, idx.stop-1):
                if self._data[i].shape[-1] != self._data[i+1].shape[0]:
                    raise ValueError(f"Core ranks do not match between adjacent cores at position {i}.")
            if norm is None:
                norm = [Normalization.NONE for _ in range(idx.stop - idx.start)]
            self._norm[idx] = norm
        else:
            raise TypeError("Invalid arguments for set_data.")

    def normalize(self, begin: int, end: Optional[int] = None) -> None:
        begin = transform_index(begin, len(self._data))
        if end is None:
            end = begin+1
        else:
            end = transform_index(end-1, len(self._data))+1
        if end < begin or end > len(self._data):
            raise IndexError("End index out of range.")

        xp = namespace_of_arrays(self._data[0])
        qr = TensorDecomposition(QRDecomposition[T]())

        for i in range(begin):
            res = qr.left(self._data[i], -1)
            self._data[i], r = res.left, res.right
            idx = len(r.shape)-1
            self._data[i+1] = xp.tensordot(r, self._data[i+1],
                                           axes=([idx], [0]))
        for i in range(len(self._data)-1, end-1, -1):
            res = qr.right(self._data[i], 1)
            l, self._data[i] =  res.left, res.right
            idx = len(self._data[i-1].shape)-1
            self._data[i-1] = xp.tensordot(self._data[i-1], l,
                                           axes=([idx], [0]))
    
    def reverse(self) -> None:
        xp = namespace_of_arrays(self._data[0])
        self._data.reverse()
        for i in range(len(self._data)):
            num = len(self._data[i].shape)
            self._data[i] = xp.permute_dims(self._data[i], range(num)[::-1])
        self._shape = self._shape.reverse()
        self._norm.reverse()

    def extend(self, train: Self, /) -> Self:
        dims = [*self.shape.dims, *train.shape.dims]
        digits = [*self.shape.digits, *train.shape.digits]
        shape = TrainShape(dims, digits)
        data = [*self.data, *train.data]
        self._base = TrainBase(shape, data)
        return self

    def _check_shape_vs_data(self, shape: TrainShape, data: Sequence[ArrayLike]) -> None:
        if len(shape) != len(data):
            raise ValueError("Number of cores in shape and data must match.")
        for i, dat in enumerate(data):
            if tuple(dat.shape[1:-1]) != tuple(shape.middle(i)):
                raise ValueError("Core data shape does not match core shape.")

    def _check_dtype_and_device(self, data: Sequence[ArrayLike]) -> None:
        dev = device(data[0])
        dtype = data[0].dtype
        for core_data in data:
            if device(core_data) != dev:
                raise ValueError("All core tensors must be on the same device.")
            if core_data.dtype != dtype:
                raise ValueError("All core tensors must have the same dtype.")

    def _check_ranks(self, data: Sequence[ArrayLike]) -> None:
        for i in range(len(data)-1):
            if data[i].shape[-1] != data[i+1].shape[0]:
                raise ValueError(f"Core ranks do not match between adjacent cores at position {i}.")
        if data[0].shape[0] != 1 or data[-1].shape[-1] != 1:
            raise ValueError("First and last cores must have rank 1 on the outside")




# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from types import NoneType
from copy import deepcopy

from .trainshape import TrainShape
from .trainbase import TrainBase
from .einsumcontraction import EinsumContraction
from .contractor import ArrayContractor, OptimizeKind
from .utils import check_operand_shapes, get_shapes
from .contractorinput import ContractorInput

class FullContractor:

    optimizer: OptimizeKind
    _contr: EinsumContraction
    _inp: NoneType | ContractorInput = None
    _expr: NoneType | ArrayContractor = None

    def __init__(self,
                 contr: EinsumContraction, 
                 optimizer: OptimizeKind = "greedy") -> None:
        if contr.result_shape is not None:
            raise ValueError("FullContractor can only be used for full contractions.")

        self._contr = deepcopy(contr)
        self.optimizer = deepcopy(optimizer)

    def __call__(self, *ops: TrainBase, expr: bool = False) -> float:
        shapes = get_shapes(*ops)
        if expr or self._inp is None:
            self.calc_expressions(*shapes)
        else:
            self._inp.check_operands(*ops)
        return self._contract(*ops)

    def calc_expressions(self, *ops: TrainShape | TrainBase) -> None:
        check_operand_shapes(self._contr.operand_shapes, get_shapes(*ops))
        self._inp = ContractorInput(*ops)
        self._expr = self._expression(*ops)

    def _expression(self, *shapes: TrainShape | TrainBase) -> ArrayContractor:
        lcontr = self._contr[0]
        eq = ",".join(str(op) for op in lcontr.operands) + f"->{lcontr.result}"
        _, tns = lcontr.get_constants(*shapes)
        expr = ArrayContractor(eq, *tns, optimizer=self.optimizer)
        return expr

    def _contract(self, *ops: TrainBase) -> float:
        if self._expr is None or self._inp is None:
            raise RuntimeError("Expression or input cannot be None here.")
        tns = self._contr[0].get_data(*ops, idx_map=self._inp.idx_map)
        res = self._expr(*tns)
        return float(res[0,0])



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from .backend import ArrayLike, namespace_of_arrays
from .trainshape import TrainShape
from .trainbase import TrainBase
from .utils import namespace_of_trains

def to_tensor[T: ArrayLike](train: TrainBase[T]) -> T:
    xp = namespace_of_trains(train)
    data = xp.ones(1, device=train.device, dtype=train.dtype)
    for tdata in train.data:
        idx = len(data.shape)-1
        data = xp.tensordot(data, tdata, axes=([idx], [0]))
    data = xp.reshape(data, data.shape[:-1])
    data = digit_back_permutation(train.shape, data)
    return data

def digit_back_permutation[T: ArrayLike](tshape: TrainShape, data: T) -> T:
    xp = namespace_of_arrays(data)
    flat = sum((list(tshape.middle(i)) for i in range(len(tshape))), start=[])
    if tuple(flat) != tuple(data.shape):
        raise ValueError("Data shape does not match train cores")
    digits = [d for dim in tshape.dims for d in dim]
    ref_digits = sum((list(digits) for digits in tshape.digits), start=[])
    perm = [ref_digits.index(d) for d in digits]

    data = xp.reshape(data, [digit.base for digits in tshape.digits for digit in digits])
    new_shape = [dim.size() for dim in tshape.dims]
    data = xp.permute_dims(data, perm)
    return xp.reshape(data, new_shape)



import unittest
from typing import Any
from itertools import product

from trainsum import TrainSum
from trainsum.backend import get_index_dtype
from utils import backends

class TestUniformGrid(unittest.TestCase):

    def setUp(self) -> None:
        self.lower = -1.0
        self.upper = 1.0
        self.trainsum = [TrainSum(backend) for backend in backends]
        self.sizes = [(120,), (256,), (20, 120), (128, 128)]

    def index_tensor(self, xp: Any, dims: tuple[int, ...]) -> Any:
        int_type = get_index_dtype(xp)
        idxs = xp.zeros((len(dims), *[size for size in dims]), dtype=int_type)
        for i, size in enumerate(dims):
            cut = [xp.newaxis] * i + [slice(None)] + [xp.newaxis] * (len(dims)-i-1)
            idxs[i,...] += xp.arange(size, dtype=int_type)[*cut]
        return idxs

    def coord_tensor(self, xp: Any, dims: tuple[int, ...]) -> Any:
        coords = xp.zeros((len(dims), *[size for size in dims]))
        for i, size in enumerate(dims):
            cut = [xp.newaxis] * i + [slice(None)] + [xp.newaxis] * (len(dims)-i-1)
            coords[i,...] += xp.linspace(self.lower, self.upper, size)[*cut]
        return coords

    def test_construction(self):
        for ts, sizes in product(self.trainsum, self.sizes):
            dims = [ts.dimension(size) for size in sizes]
            domains = [ts.domain(self.lower, self.upper) for _ in sizes]
            spacings = [(domain.diff / (dim.size()-1)) for dim, domain in zip(dims, domains)]
            grid = ts.uniform_grid(dims, domains)

            self.assertTrue(all(d1==d2 for d1, d2 in zip(grid.dims, dims)))
            self.assertTrue(all(d1==d2 for d1, d2 in zip(grid.domains, domains)))
            self.assertTrue(all(d1==d2 for d1, d2 in zip(grid.spacings, spacings)))

        for ts in self.trainsum:
            with self.assertRaises(ValueError):
                dims = (ts.dimension(10),)
                domains = (ts.domain(1.0, 2.0), ts.domain(1.0, 2.0))
                grid = ts.uniform_grid(dims, domains)

    def test_to_idxs(self):
        for (xp, ts), sizes in product(zip(backends, self.trainsum), self.sizes):

            ref_idxs = self.index_tensor(xp, sizes)
            ref_coords = self.coord_tensor(xp, sizes)

            dims = [ts.dimension(size) for size in sizes]
            domains = [ts.domain(self.lower, self.upper) for _ in sizes]
            grid = ts.uniform_grid(dims, domains)
            idxs = grid.to_idxs(ref_coords)

            self.assertTrue(xp.all(idxs == ref_idxs))
            self.assertRaises(ValueError, grid.to_idxs, xp.zeros((10, 0)))

    def test_to_coords(self):
        for (xp, ts), sizes in product(zip(backends, self.trainsum), self.sizes):

            ref_idxs = self.index_tensor(xp, sizes)
            ref_coords = self.coord_tensor(xp, sizes)

            dims = [ts.dimension(size) for size in sizes]
            domains = [ts.domain(self.lower, self.upper) for _ in sizes]
            grid = ts.uniform_grid(dims, domains)
            coords = grid.to_coords(ref_idxs)

            self.assertTrue(xp.max(xp.abs(coords-ref_coords)) < 1e-5)
            self.assertRaises(ValueError, grid.to_coords, xp.zeros((10, 0), dtype=xp.int64))
    
if __name__ == '__main__':
    unittest.main()



import unittest

from trainsum import TrainSum
from trainsum.typing import OptionType
from utils import backends

class TestTensorTrain(unittest.TestCase):

    def setUp(self):
        self.trainsum = [TrainSum(backend) for backend in backends]

    def test_options(self) -> None:

        opts = [(lambda ts: ts.exact(), OptionType.EINSUM),
                (lambda ts: ts.decomposition(max_rank=10), OptionType.EINSUM),
                (lambda ts: ts.variational(max_rank=10), OptionType.EINSUM),
                (lambda ts: ts.evaluation(), OptionType.EVALUATE),
                (lambda ts: ts.cross(max_rank=10, eps=1e-10), OptionType.CROSS)]

        for ts in self.trainsum:

            for (ofunc, otype) in opts:
                with ofunc(ts) as opts1:
                    with ofunc(ts) as opts2:
                        self.assertEqual(opts2, ts.get_options(otype))
                    self.assertEqual(opts1, ts.get_options(otype))

                opt = ofunc(ts)
                ts.set_options(opt)
                self.assertEqual(opt, ts.get_options(otype))

if __name__ == "__main__":
    unittest.main()



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from .digit import Digit
from .dimension import Dimension
from .domain import Domain
from .uniformgrid import UniformGrid
from .trainshape import TrainShape, trainshape

from .matrixdecomposition import MatrixDecomposition
from .qrdecomposition import QRDecomposition
from .svdecomposition import SVDecomposition
from .sweepingstrategy import SweepingStrategy

from .options import set_options, get_options, OptionType



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from types import NoneType
from dataclasses import dataclass
from collections import deque
from copy import deepcopy

from .backend import ArrayLike, shape
from .digit import Digits
from .direction import Direction
from .trainshape import TrainShape
from .trainbase import TrainBase
from .matrixdecomposition import MatrixDecompositionResult
from .tensordecomposition import TensorDecomposition
from .utils import get_device_dtype, namespace_of_trains, block_tensor, sequence_product

@dataclass(kw_only=True)
class CutResult[T: ArrayLike]:
    digits: Digits
    data: T

class AdditionCore[T: ArrayLike]:

    @property
    def data(self) -> T:
        if self._data is None:
            raise RuntimeError("AdditionCore is empty, cannot return data.")
        return self._data

    _data: NoneType | T
    _direc: Direction
    _digit_queue: deque[Digits]
    _shape: TrainShape

    def __init__(self, shape: TrainShape, direction: Direction) -> None:
        self._shape = deepcopy(shape)
        self._direc = direction
        self._digit_queue = deque()
        self._data = None
    
    def add(self, idx: int, *trains: TrainBase[T]) -> None:
        #if not all(self._shape[idx].similar(train.shape[idx]) for train in trains):
        #    raise ValueError("Train shapes do not match for addition at index {idx}.")
        digits = self._shape.digits[idx]
        if self._direc == Direction.TO_RIGHT:
            self._digit_queue.append(digits)
        else: # Direction.TO_LEFT
            self._digit_queue.appendleft(digits)

        if self._data is None:
            self._data = self._add_first(idx, *trains)
        elif self._direc == Direction.TO_RIGHT:
            self._data = self._add_right(idx, *trains)
        else: # Direction.TO_LEFT
            self._data = self._add_left(idx, *trains)

    def cut[S: MatrixDecompositionResult](self, decomp: TensorDecomposition[T, S]) -> CutResult[T]:
        if self._data is None:
            raise RuntimeError("Cannot cut when data is None.")
        first_fit = all(d1==d2 for d1, d2 in zip(self._digit_queue[0], self._shape.digits[0]))
        last_fit  = all(d1==d2 for d1, d2 in zip(self._digit_queue[-1], self._shape.digits[-1]))
        if len(self._digit_queue) == 1 and (first_fit or last_fit):
            data = self._data
            self._data = None
            digits = self._digit_queue.pop()
        elif self._direc == Direction.TO_RIGHT:
            digits = self._digit_queue.popleft()
            num = 1+len(digits)
            res = decomp.left(self._data, num)
            self._data = res.right
            data = res.left
        else: # Direction.TO_LEFT
            digits = self._digit_queue.pop()
            num = 1+len(digits)
            res = decomp.right(self._data, -num)
            self._data = res.left
            data = res.right
        return CutResult(digits=digits, data=data)

    def _add_first(self, idx: int, *trains: TrainBase[T]) -> T:
        xp = namespace_of_trains(*trains)
        device, dtype = get_device_dtype(trains)

        tns = [train.data[idx] for train in trains]

        left = 0 if idx == 0 else slice(None)
        left_size = 1 if idx == 0 else sum(shape(tn)[0] for tn in tns)
        right = 0 if idx == len(trains[0].shape)-1 else slice(None)
        right_size = 1 if idx == len(trains[0].shape)-1 else sum(shape(tn)[-1] for tn in tns)
        middle = self._shape.middle(idx)

        data = xp.zeros((left_size, *middle, right_size),
                        device=device, dtype=dtype)
        for idxs in sequence_product(self._shape.middle(idx)):
            cut = (left, *idxs, right)
            data[cut] = block_tensor(*[tn[cut] for tn in tns])

        return data

    def _add_left(self, idx: int, *trains: TrainBase[T]) -> T:
        if self._data is None:
            raise RuntimeError("Cannot add left when data is None.")

        xp = namespace_of_trains(*trains)
        device, dtype = get_device_dtype(trains)
        is_begin = (idx == 0)

        tns = [train.data[idx] for train in trains]
        shapes = [shape(tn) for tn in tns]
        dshape = shape(self._data)

        left = 1 if is_begin else sum(sh[0] for sh in shapes)
        middle = self._shape.middle(idx)
        if sum(sh[-1] for sh in shapes) != dshape[0]:
            raise ValueError("Left ranks do not match for addition.")

        data = xp.zeros((left, *middle, *dshape[1:]),
                        device=device, dtype=dtype)

        off0, off1 = 0, 0
        idx = len(shapes[0]) - 1
        for sh, tn in zip(shapes, tns):
            cut0 = slice(off0, off0+sh[-1])
            cut1 = slice(None) if is_begin else slice(off1, off1+sh[0])
            data[cut1,...] += xp.tensordot(tn, self._data[cut0,...], axes=([idx],[0]))
            off0 += sh[-1]
            off1 += sh[0]

        return data

    def _add_right(self, idx: int, *trains: TrainBase[T]) -> T:
        if self._data is None:
            raise RuntimeError("Cannot add right when data is None.")

        xp = namespace_of_trains(*trains)
        device, dtype = get_device_dtype(trains)
        is_end = (idx == len(self._shape)-1)

        tns = [train.data[idx] for train in trains]
        shapes = [shape(tn) for tn in tns]
        dshape = shape(self._data)

        middle = self._shape.middle(idx)
        right = 1 if is_end else sum(shape[-1] for shape in shapes)
        if sum(sh[0] for sh in shapes) != dshape[-1]:
            raise ValueError("Right ranks do not match for addition.")

        data = xp.zeros((*dshape[:-1], *middle, right),
                        device=device, dtype=dtype)

        off0, off1 = 0, 0
        idx = len(dshape) - 1
        for sh, tn in zip(shapes, tns):
            cut0 = slice(off0, off0+sh[0])
            cut1 = slice(None) if is_end else slice(off1, off1+sh[-1])
            data[...,cut1] += xp.tensordot(self._data[...,cut0], tn, axes=([idx],[0]))
            off0 += sh[0]
            off1 += sh[-1]

        return data



# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

from math import prod
from typing import Optional, Sequence, Callable, Generator
from collections import deque
from copy import deepcopy
from random import randint

from .backend import ArrayLike, namespace_of_arrays, get_index_dtype, shape, ArrayNamespace
from .digit import Digit
from .dimension import Dimension
from .trainshape import TrainShape
from .trainbase import TrainBase
from .localrange import LocalRange
from .sweepingstrategy import SweepingStrategy
from .utils import namespace_of_trains
from .matrixleastsquares import MatrixLeastSquares
from .full import full

from .functionalcore import FunctionalCore
from .picore import PiCore

class CrossInterpolation[T: ArrayLike]:

    solver: MatrixLeastSquares
    eps: float
    _strat: SweepingStrategy

    @property
    def strategy(self) -> SweepingStrategy:
        return self._strat
    @strategy.setter
    def strategy(self, strat: SweepingStrategy) -> None:
        if strat.ncores != 2 or not strat.min_size is None:
            raise NotImplementedError("Only SweepingStrategy with ncores=2 and min_size=0 is supported")
        self._strat = strat

    def __init__(
            self,
            solver: MatrixLeastSquares,
            strategy: SweepingStrategy = SweepingStrategy(ncores=2, nsweeps=10),
            eps: float = 1e-6) -> None:
        self.solver = solver
        self.strategy = strategy
        self.eps = eps

    def __call__(
            self,
            xp: ArrayNamespace[T],
            func: Callable[[T], T],
            tshape: TrainShape,
            start_idxs: Optional[T] = None,
            ) -> TrainBase[T]:
        train = full(xp, tshape, 0.0)
        for lrange, eps in self._gen(func, train, start_idxs):
            #print(f"eps={eps:.2e} at range {lrange.begin}:{lrange.end}", end='\n', flush=True)
            pass
        return train

    def _gen(
            self,
            func: Callable[[T], T],
            train: TrainBase[T],
            start_idxs: Optional[T] = None,
            ) -> Generator[tuple[LocalRange, float]]:

        max_rank = 2*self._strat.nsweeps+1
        cross_gen = cross_interpolation(func, train, self.solver, max_rank, start_idxs)
        eps = deepcopy(self.eps)
        try:
            errs = deque[float]()
            for _, lrange in self._strat(train.shape):
                err_val = cross_gen.send(lrange)
                errs.append(err_val)
                if len(errs) == len(train.shape):
                    errs.popleft()

                mean = sum(errs)/len(errs)
                yield lrange, mean
                if mean < eps:
                    break
            raise GeneratorExit

        except GeneratorExit:
            cross_gen.close()

    def _get_strat(self, shape: TrainShape) -> Generator[tuple[int, LocalRange]]:
        right1 = [LocalRange(begin=i, end=i+2) for i in range(0, len(shape), 2) if i+2 <= len(shape)]
        right2 = [LocalRange(begin=i, end=i+2) for i in range(1, len(shape), 2) if i+2 <= len(shape)]
        for i in range(4*self._strat.nsweeps):
            if i%2 == 0:
                for j in range(len(right1)):
                    yield i//2, right1[j]
            else:
                for j in range(len(right2)):
                    yield i//2, right2[j]

def cross_interpolation[T: ArrayLike](
        func: Callable[[T], T],
        train: TrainBase[T],
        solver: MatrixLeastSquares,
        max_rank: int,
        start_idxs: Optional[T] = None,
        ) -> Generator[float, LocalRange]:
    gen = cross_interpolation_gen(func, train, solver, max_rank, start_idxs)
    next(gen)
    return gen

def cross_interpolation_gen[T: ArrayLike](
        func: Callable[[T], T],
        train: TrainBase[T],
        solver: MatrixLeastSquares,
        max_rank: int,
        start_idxs: Optional[T] = None,
        ) -> Generator[float, LocalRange]:

    def func_(x: T) -> T:
        xp = namespace_of_arrays(x)
        x_ = xp.reshape(x, (x.shape[0], prod(shape(x)[1:])))
        res = func(x_)
        return xp.reshape(res, x.shape[1:])

    xp = namespace_of_trains(train)
    int_type = get_index_dtype(xp)
    dims, digits = train.shape.dims, train.shape.digits
    idx_array = lambda sh: xp.empty((len(dims), sh),
                                     device=train.device,
                                     dtype=int_type)

    if start_idxs is None:
        sidxs = xp.asarray(random_idxs(dims), dtype=int_type)[:,xp.newaxis]
    else:
        sidxs = xp.astype(start_idxs, int_type)

    fcores = [FunctionalCore(dims, digits[0], idx_array(1), idx_array(max_rank))]
    for dgts in digits[1:-1]:
        fcores.append(FunctionalCore(dims, dgts, idx_array(max_rank), idx_array(max_rank)))
    fcores.append(FunctionalCore(dims, digits[-1], idx_array(max_rank), idx_array(1)))
    for i in range(shape(sidxs)[1]):
        idxs = xp.reshape(sidxs[:,i], (sidxs.shape[0], 1))
        add_idxs(dims, fcores, idxs)

    picores = [PiCore(fcores[i], fcores[i+1], func_, train.dtype) for i in range(len(fcores)-1)]

    err_val = xp.inf
    try:
        while True:
            data = yield err_val
            if data is None:
                raise ValueError("Expected SweepingData, got None")
            elif data.end-data.begin != 2:
                raise ValueError(f"Expected SweepingData with ncores=2, got {data}")

            lcore, rcore, picore = fcores[data.begin], fcores[data.end-1], picores[data.begin]
            picore.update_rows(lcore, rcore, func_)
            picore.update_cols(lcore, rcore, func_)
            err = picore.get_error(solver)

            idxs = get_index(xp.argmax(err), err.shape)
            left = (lcore.left.data[:,idxs[0]] + lcore.middle.data[:,idxs[1]])
            right = (rcore.middle.data[:,idxs[2]] + rcore.right.data[:,idxs[3]])

            lcore.right.add(right)
            picore.set_col(right)
            rcore.left.add(left)
            picore.set_row(left)

            err_val = float(xp.linalg.vector_norm(err) / xp.linalg.vector_norm(picore._data[...]))

    except GeneratorExit:
        for i in range(len(picores)-1):
            lcore, rcore, picore = fcores[i], fcores[i+1], picores[i]
            picore.update_rows(lcore, rcore, func_)
            picore.update_cols(lcore, rcore, func_)

        data = [picores[0].left()]
        for i in range(1, len(fcores)):
            data.append(picores[i-1].result_right(solver))
        train.set_data(slice(0, len(data)), data)

def random_idxs(dims: Sequence[Dimension]) -> list[int]:
    return [randint(0, dim.size()-1) for dim in dims]

def add_idxs[T: ArrayLike](dims: Sequence[Dimension], fcores: Sequence[FunctionalCore[T]], idxs: T) -> None:
    xp = namespace_of_arrays(idxs)
    int_type = get_index_dtype(xp)
    digits = [dim.to_digits(idxs[i]) for i, dim in enumerate(dims)]
    didxs = [dim_idxs(dims, fcore) for fcore in fcores]

    for i, fcore in enumerate(fcores):

        tmp = xp.zeros((len(dims)), device=fcore.device, dtype=int_type)
        for j in range(i):
            tmp += eval_middle(fcores[j], didxs[j], digits)
        tmp = xp.reshape(tmp, (tmp.shape[0], 1))
        fcore.left.add(tmp)

        tmp = xp.zeros((len(dims)), device=fcore.device, dtype=int_type)
        for j in range(len(fcores)-1, i, -1):
            tmp += eval_middle(fcores[j], didxs[j], digits)
        tmp = xp.reshape(tmp, (tmp.shape[0], 1))
        fcore.right.add(tmp)

def eval_middle[T: ArrayLike](fcore: FunctionalCore[T], dim_idxs: Sequence[int], digits: Sequence[T]) -> T:
    middle_idxs = [int(digits[dim_idx][digit.idx][0]) for dim_idx, digit in zip(dim_idxs, fcore)]
    view = [fcore.middle.data.shape[0]] + [digit.base for digit in fcore]
    xp = namespace_of_arrays(fcore.middle.data)
    return xp.reshape(fcore.middle.data, view)[:,*middle_idxs]

def dim_idxs(dims: Sequence[Dimension], digits: Sequence[Digit]) -> Sequence[int]:
    idfs = [dim.idf for dim in dims]
    return [idfs.index(digit.idf) for digit in digits]

def get_index(idx: int, shape: Sequence[int]) -> Sequence[int]:
    res = []
    for dim in reversed(shape):
        res.append(int(idx % dim))
        idx //= dim
    return list(reversed(res))
